# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

Binning!

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings

import numpy
from scipy import sparse

from astropy.io import fits

from vorbin.voronoi_2d_binning import voronoi_2d_binning

from ..par.parset import KeywordParSet
from ..util.geometry import SemiMajorAxisCoo
from ..util.covariance import Covariance


# BASE CLASS -----------------------------------------------------------

class SpatialBinning():
    """
    Base class for spatially binning data.
    """
    def __init__(self, bintype, par=None):
        self.bintype = bintype
        self.par = par

    @staticmethod
    def bin_index(x, y, gpm=None, par=None):
        """
        Undefined in base class. Just provides expected calling
        sequence.
        """
        pass
# ----------------------------------------------------------------------


# GLOBAL BINNING -------------------------------------------------------
class GlobalBinning(SpatialBinning):
    """
    Class that performs the global binning.
    """
    def __init__(self):
        SpatialBinning.__init__(self, 'global')

    @staticmethod
    def bin_index(x, y, gpm=None, par=None):
        """
        Bin the data and return the indices of the bins.

        The calling sequence is just needed to match the expectation
        for a :class:`SpatialBinning` object.

        Args:
            x (`numpy.ndarray`_):
                Fiducial on-sky X position of each spectrum. X
                increases with RA. Only used to set the size of the
                returned array.
            y (`numpy.ndarray`_):
                **Not used.**
            gpm (`numpy.ndarray`_, optional):
                Good value (pixel) mask used to select data that should be
                included in the bin.  If None, all values are considered valid.
                If provided, masked values (gpm=False) are given a bin index of
                -1.
            par (object, optional):
                **Not used.**

        Returns:
            `numpy.ndarray`_: An integer bin index for each spectrum.
        """
        binid = numpy.zeros(x.shape, dtype=int)
        if gpm is not None:
            if gpm.size != x.size:
                raise ValueError(f'Provided GPM has incorrect shape; found {gpm.size}, ' \
                                 f'expected {x.size}.')
            binid[numpy.logical_not(gpm)] = -1
        return binid
# ----------------------------------------------------------------------

# RADIAL BINNING -------------------------------------------------------
class RadialBinningPar(KeywordParSet):
    """
    Class with parameters used by the radial binning algorithm.  See
    :class:`mangadap.par.parset.ParSet` for attributes.

    The defined parameters are:

    .. include:: ../tables/radialbinningpar.rst
    """
    def __init__(self, center=[0., 0.], pa=-1, ell=-1, radius_scale=-1, radii=[-1,-1,10],
                 log_step=False):
        in_fl = [ int, float ]
        ar_like = [ numpy.ndarray, list ]
        
        pars =   [ 'center',  'pa', 'ell', 'radius_scale', 'radii', 'log_step' ]
        values = [   center,    pa,   ell,   radius_scale,   radii,   log_step ]
        dtypes = [  ar_like, in_fl, in_fl,          in_fl, ar_like,       bool ]

        descr = ['A two-element array defining the center to use in the definition of the ' \
                    'elliptical bins.  This is defined as a sky-right offset in arcseconds from ' \
                    'the nominal center of the object.',
                 'Sets the position angle, defined from N through E of the major axis of the ' \
                    'isophotal ellipse used to define the elliptical bins.  Set to <0 to ' \
                    'indicate that the value should be filled by the metadata in the reduction ' \
                    'assessment class; see :func:`fill`.',
                 'Sets the ellipticity (1-b/a) of the isophotal ellipse use to define the ' \
                    'elliptical bins.  Set to <0 to indicate that the value should be filled by ' \
                    'the metadata in the reduction assessment class; see :func:`fill`.',
                 'Defines a scale factor to use when defining the radial bins.  For example, ' \
                    'you might want to scale to the a certain number of effective radii or ' \
                    'physical scale in kpc.  For no scale, use 1.0.  Set to <0 to indicate that ' \
                    'the value should be filled by the metadata in the reduction assessment ' \
                    'class; see :func:`fill`.',
                 'A three-element array defining the starting and ending radius for the bin ' \
                    'edges and the number of bins to create.  If the starting radius is -1, ' \
                    'the inner-most radius is set to 0 when not using log bins or 0.1 arcsec ' \
                    'when using logarithmically spaced bins.  If the ending radius is -1, the ' \
                    'outer-most radius is set by the spaxel at the largest radius.',
                 'A flag that the radial bins should be a geometric series.']

        super().__init__(pars, values=values, dtypes=dtypes, descr=descr)
    
    def toheader(self, hdr):
        """
        Copy some of the parameters to a header.
        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to write to.

        Returns:
            `astropy.io.fits.Header`_: Edited header object

        Raises:
            TypeError:
                Raised if input is not an `astropy.io.fits.Header`_
                object.
        """
        if not isinstance(hdr, fits.Header):
            raise TypeError('Input is not a astropy.io.fits.Header object!')

        hdr['BINCX'] = (self['center'][0], 'Radial binning on-sky X center')
        hdr['BINCY'] = (self['center'][1], 'Radial binning on-sky Y center')
        hdr['BINPA'] = (self['pa'], 'Radial binning position angle')
        hdr['BINELL'] = (self['ell'], 'Radial binning ellipticity (1-b/a)')
        hdr['BINSCL'] = (self['radius_scale'], 'Radial binning radius scale (arcsec)')
        hdr['BINRAD'] = (str(list(self['radii'])), 'Radial binning radius sampling')
        hdr['BINLGR'] = (str(self['log_step']), 'Geometric step used by radial binning')
        return hdr

    def fromheader(self, hdr):
        """
        Copy the information from the header.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to read from.
        """
        if not isinstance(hdr, fits.Header):
            raise TypeError('Input is not a astropy.io.fits.Header object!')

        self['center'] = [ hdr['BINCX'], hdr['BINCY'] ]
        self['pa'] = hdr['BINPA']
        self['ell'] = hdr['BINELL']
        self['radius_scale'] = hdr['BINSCL']
        self['radii'] = eval(hdr['BINRAD'])
        self['log_step'] = bool(hdr['BINLGR'])

    # TODO: This should actually use the cube metadata, not the rdxqa
    def fill(self, rdxqa):
        """
        Use the metadata in the reduction assessment results to fill out the
        binning parameters.
        """
        if self['pa'] < 0:
            self['pa'] = 0. if rdxqa.pa is None else rdxqa.pa
        if self['ell'] < 0:
            self['ell'] = 0. if rdxqa.ell is None else rdxqa.ell
        if self['radius_scale'] < 0:
            self['radius_scale'] = 1.0 if rdxqa.reff is None else rdxqa.reff


class RadialBinning(SpatialBinning):
    """
    Class that performs the radial binning.
    """
    def __init__(self, par=None):
        SpatialBinning.__init__(self, 'radial', par=par)
        self.sma_coo = None
        self.to_center = False
        self.rs = None
        self.dr = None

    def _r_start_step(self, r):
        """
        Bin starting radii are::
            
            rs = self.par['radii'][0]
            re = self.par['radii'][1]
            nr = int(self.par['radii'][2])
            starting_radii = numpy.logspace(numpy.log10(rs), numpy.log10(re), num=nr,
                                            endpoint=False) if self.par['log_step'] else \
                             numpy.linspace(rs, re, num=nr, endpoint=False)

        For ending radii, just swap ``rs`` and ``re`` above.

        Args:
            r (`numpy.ndarray`_):
                Fiducial radius of each spectrum.
        """
        # Set minimum radius, if necessary
        if self.par['radii'][0] < 0:
            ndec = int(numpy.floor(numpy.absolute(numpy.log10(0.5/self.par['radius_scale']))))
            ndec = 2 if ndec < 2 else ndec
            self.par['radii'][0] = numpy.around(0.5/self.par['radius_scale'], decimals=ndec) \
                                        if self.par['log_step'] else 0.0
            self.to_center = True

        # Set maximum radius, if necessary
        if self.par['radii'][1] < 0:
            self.par['radii'][1] = numpy.around(numpy.amax(r)+0.1, decimals=2)

        # Get the logarithmically stepped values
        if self.par['log_step']:
            self.rs = numpy.log10(self.par['radii'][0])
            self.dr = (numpy.log10(self.par['radii'][1]) - self.rs)/self.par['radii'][2]
            return

        # Get the linearly stepped values
        self.rs = self.par['radii'][0]
        self.dr = (self.par['radii'][1] - self.rs)/self.par['radii'][2]
        
    def bin_index(self, x, y, gpm=None, par=None):
        """
        Bin the data and return the indices of the bins.

        Args:
            x (`numpy.ndarray`_):
                Fiducial on-sky X position of each spectrum. X
                increases with RA.
            y (`numpy.ndarray`_):
                Fiducial on-sky Y position of each spectrum. Y
                increases with DEC.
            gpm (`numpy.ndarray`_, optional):
                Good value (pixel) mask used to select data that should be
                included in the bin.  If None, all values are considered valid.
                If provided, masked values (gpm=False) are given a bin index of
                -1.
            par (:class:`RadialBinningPar`, optional):
                Binning parameters.  Cannot be None.

        Returns:
            `numpy.ndarray`_: An integer bin index for each spectrum.

        Raises:
            ValueError:
                Raised if the sizes of ``x`` and ``y`` do not match
                or if ``par`` is None.
        """
        # Check the input
        if x.size != y.size:
            raise ValueError('Dimensionality of x and y coordinates do not match!')
        if gpm is not None and gpm.size != x.size:
            raise ValueError(f'Provided GPM has incorrect shape; found {gpm.size}, ' \
                             f'expected {x.size}.')

        # Perform any necessary initialization
        if par is not None:
            self.par = par
            if self.sma_coo is not None:
                del self.sma_coo
                self.sma_coo = None
        if self.par is None:
            raise ValueError('Required parameters for RadialBinning have not been defined.')
        if self.sma_coo is None:
            self.sma_coo = SemiMajorAxisCoo(xc=self.par['center'][0], yc=self.par['center'][1],
                                            rot=0.0, pa=self.par['pa'], ell=self.par['ell'])

        # Get the radius and azimuth
        r, theta = self.sma_coo.polar(x,y)

        # Get the starting radius and step per bin
        r /= self.par['radius_scale']
        indx = r >= 0.0
        if self.par['log_step']:
            indx = r > 0
        self._r_start_step(r[indx])

        # Determine the bin indices
        if self.par['log_step']:
            r[indx] = numpy.log10(r[indx])

        binid = numpy.floor((r - self.rs)/self.dr).astype(int)
        if self.to_center:
            binid[(binid<0) & indx] = 0          # Go all the way to R=0
        else:
            binid[r<self.rs] = -1            # or not
        binid[binid >= int(self.par['radii'][2])] = -1  # Remove any points outside the last bin

        if gpm is not None:
            # Remove any masked pixels
            binid[gpm] = -1

        return binid

    def bin_area(self):
        """
        Return the nominal area for each elliptical bin.

        Returns:
            `numpy.ndarray`_: Array with the analytic on-sky area of
            each elliptical bin in square arcsec given its radial
            limits.

        Raises:
            ValueError:
                Raised if :attr:`par` is None.
        """
        if self.par is None:
            raise ValueError('Required parameters not defined.')

        rs = self.par['radii'][0]
        re = self.par['radii'][1]
        nr = int(self.par['radii'][2])
        bin_radii = numpy.append(
                        numpy.logspace(numpy.log10(rs), numpy.log10(re), num=nr, endpoint=False) \
                        if self.par['log_step'] \
                        else numpy.linspace(rs, re, num=nr, endpoint=False), re) \
                                * self.par['radius_scale']

        return numpy.pi*(1.0-self.par['ell'])*numpy.diff(numpy.square(bin_radii))
# ----------------------------------------------------------------------


# VORONOI BINNING ------------------------------------------------------
class VoronoiBinningPar(KeywordParSet):
    r"""
    Class with parameters used by the Voronoi binning algorithm.

    See :class:`mangadap.par.parset.ParSet` for attributes.  See
    `vorbin`_ for the main algorithm.

    The defined parameters are:

    .. include:: ../tables/voronoibinningpar.rst
    """
    def __init__(self, target_snr=10., signal=None, noise=None, covar=None):
        in_fl = [ int, float ]
        covar_type = [ float, numpy.ndarray, Covariance, sparse.spmatrix ]
        ar_like = [ numpy.ndarray, list ]
        
        pars =   [ 'target_snr', 'signal', 'noise',    'covar' ]
        values = [   target_snr,   signal,   noise,      covar ]
        dtypes = [        in_fl,  ar_like, ar_like, covar_type ]

        descr = ['The target S/N for each bin.',
                 'The array of signal measurements for each on-sky position to bin.  See ' \
                    ':func:`fill` to fill this based on the data in the reduction assessments ' \
                    'object.',
                 'The array of noise measurements for each on-sky position to bin.  If not ' \
                    'provided, ``covar`` must be provided and be a full covariance matrix.  See ' \
                    ':func:`fill` to fill this based on the data in the reduction assessments ' \
                    'object.',
                 r'Covariance matrix or calibration normalization.  For the latter, the value ' \
                    r'is used to renormalize using :math:`n_{\rm calib} = n_{\rm nominal} ' \
                    r'(1 + \alpha\ \log\ N_{\rm bin})`, where :math:`N_{\rm bin}` is the number ' \
                    r'of binned spaxels and :math:`\alpha` is the value provided. See ' \
                    r':func:`fill` to fill this based on the data in the reduction assessments ' \
                    r'object.']

        super().__init__(pars, values=values, dtypes=dtypes, descr=descr)

    def toheader(self, hdr):
        """
        Copy some of the parameters to a header.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to write to.

        Returns:
            `astropy.io.fits.Header`_: Edited header object

        Raises:
            TypeError:
                Raised if input is not an `astropy.io.fits.Header`_
                object.
        """
        if not isinstance(hdr, fits.Header):
            raise TypeError('Input is not a astropy.io.fits.Header object!')

        hdr['BINSNR'] = (self['target_snr'], 'Voronoi binning minimum S/N')
        if isinstance(self['covar'], float):
            hdr['BINCOV'] = ('calib', 'Voronoi binning S/N covariance type')
            hdr['NCALIB'] = (self['covar'], 'Noise calibration value')
        elif isinstance(self['covar'], (numpy.ndarray, Covariance, sparse.spmatrix)):
            hdr['BINCOV'] = ('full', 'Voronoi binning S/N covariance type')
        else:
            hdr['BINCOV'] = ('none', 'Voronoi binning S/N covariance type')
        return hdr

    def fromheader(self, hdr):
        """
        Copy the information from the header.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to read from.
        """
        if not isinstance(hdr, fits.Header):
            raise TypeError('Input is not a astropy.io.fits.Header object!')

        self['target_snr'] = hdr['BINSNR']
        if hdr['BINCOV'] == 'calib':
            self['covar'] = hdr['NCALIB']

    def fill(self, rdxqa):
        """
        Use the data in the reduction assessment results to fill out the binning
        parameters.
        """
        self['signal'] = rdxqa['SPECTRUM'].data['SIGNAL']
        if rdxqa.correlation is None:
            self['noise'] = numpy.sqrt(rdxqa['SPECTRUM'].data['VARIANCE'])
            return
        # Overwrite any existing calibration coefficient
        rdxqa.correlation.revert_correlation()
        covar = rdxqa.correlation.toarray()
        rdxqa.correlation.to_correlation()
        i, j = numpy.meshgrid(numpy.arange(covar.shape[0]), numpy.arange(covar.shape[1]))
        self['covar'] = Covariance(inp=sparse.coo_matrix((covar[covar > 0].ravel(),
                                   (i[covar > 0].ravel(), j[covar > 0].ravel())),
                                   shape=covar.shape).tocsr())


class VoronoiBinning(SpatialBinning):
    """
    Class that wraps the contributed voronoi binning code.
    """
    def __init__(self, par=None):
        SpatialBinning.__init__(self, 'voronoi', par=par)
        self.covar = None

    def sn_calculation_no_covariance(self, index, signal, noise): 
        """
        S/N calculation built to match the requirements of
        :func:`mangadap.contrib.voronoi_2d_binning`.

        This calculation is identical to the one provided directly by
        the code.

        Args:
            index (`numpy.ndarray`_):
                Indices of the measurements in a single bin.
            signal (`numpy.ndarray`_):
                The signal measurements.
            noise (`numpy.ndarray`_):
                The noise measurements.

        Returns:
            :obj:`float`: The nominal signal-to-noise reached by
            summing the measurements selected by ``index``.
        """
        return  numpy.sum(signal[index]) / numpy.sqrt(numpy.sum(numpy.square(noise[index])))

    def sn_calculation_covariance_matrix(self, index, signal, noise):
        """
        Calculate the S/N using a full covariance matrix.

        The method uses the internal :attr:`covar`.

        Args:
            index (`numpy.ndarray`_):
                Indices of the measurements in a single bin.
            signal (`numpy.ndarray`_):
                The signal measurements.
            noise (`numpy.ndarray`_):
                The noise measurements.

        Returns:
            :obj:`float`: The nominal signal-to-noise reached by
            summing the measurements selected by ``index``, including
            any covariance.
        """
        _index = numpy.atleast_1d(index)
        return numpy.sum(signal[index])/numpy.sqrt(numpy.sum(self.covar[_index,:][:,_index]))

    def sn_calculation_calibrate_noise(self, index, signal, noise):
        r"""
        Calculate the S/N using a calibration of the S/N following:

        .. math::

            N_{\rm calib} = N_{\rm nominal} (1 + \alpha\ \log N_{\rm bin})

        where :math:`N_{\rm bin}` is the number of binned spaxels and
        :math:`\alpha` is an empirically derived constant used to adjust
        the noise for the affects of covariance.

        The calibration constant, :math:`\alpha`, is kept internally
        using :attr:`covar`.

        Args:
            index (`numpy.ndarray`_):
                Indices of the measurements in a single bin.
            signal (`numpy.ndarray`_):
                The signal measurements.
            noise (`numpy.ndarray`_):
                The noise measurements.

        Returns:
            :obj:`float`: The nominal signal-to-noise reached by
            summing the measurements selected by ``index``, after
            applying the covariance calibration.
        """
        return numpy.sum(signal[index]) / (numpy.sqrt(numpy.sum(numpy.square(noise[index]))) \
                        * (1.0 + self.covar*numpy.log10(len(signal[index]))))

    def bin_index(self, x, y, gpm=None, par=None):
        """
        Bin the data and return the indices of the bins.

        Args:
            x (`numpy.ndarray`_):
                Fiducial on-sky X position of each spectrum. X
                increases with RA.
            y (`numpy.ndarray`_):
                Fiducial on-sky Y position of each spectrum. Y
                increases with DEC.
            gpm (`numpy.ndarray`_, optional):
                Good value (pixel) mask used to select data that should be
                included in the bin.  If None, all values are considered valid.
                If provided, masked values (gpm=False) are given a bin index of
                -1.
            par (:class:`RadialBinningPar`, optional):
                Binning parameters.  Cannot be None.

        Returns:
            `numpy.ndarray`_: An integer bin index for each spectrum.

        Raises:
            ValueError:
                Raised if the sizes of ``x`` and ``y`` do not match,
                if ``par`` is None, if various checks of the signal,
                noise, or covariance elements are incorrectly
                matched.
        """
        # Check the position input
        if x.size != y.size:
            raise ValueError('Dimensionality of x and y coordinates do not match!')
        if gpm is not None and gpm.size != x.size:
            raise ValueError(f'Provided GPM has incorrect shape; found {gpm.size}, ' \
                             f'expected {x.size}.')

        # Initialize the parameter object
        if par is not None:
            self.par = par
        if self.par is None:
            raise ValueError('Required parameters for Voronoi binning have not been defined.')

        # Check the signal input
        if self.par['signal'] is None:
            raise ValueError('Signal measurements not provided for Voronoi binning!')
        if self.par['signal'].size != x.size:
            raise ValueError('Dimensionality of signal does not match on-sky coordinates.')

        # Construct the covariance input if provided
        self.covar = None
        sn_func = self.sn_calculation_no_covariance
        if self.par['covar'] is not None:
            if isinstance(self.par['covar'], Covariance):
                # Check dimensionality
                if self.par['covar'].dim == 3:
                    raise ValueError('Input Covariance object must be two-dimensional')
                # Make sure the object is a full covariance matrix
                if self.par['covar'].is_correlation:
                    self.par['covar'].revert_correlation()
            # Fill the full array if necessary
            if not isinstance(self.par['covar'], (numpy.ndarray,float)):
                self.covar = self.par['covar'].toarray()
                sn_func = self.sn_calculation_covariance_matrix
            else:
                self.covar = self.par['covar']
                sn_func = self.sn_calculation_calibrate_noise

            if isinstance(self.covar, numpy.ndarray):
                # Check that the size matches the input signal array
                if self.covar.shape[0] != self.par['signal'].size:
                    raise ValueError('Dimensionality of covariance does not match signal.')
                _noise = numpy.sqrt(numpy.diag(self.covar))

        # Make sure the noise vector is available
        if self.covar is None or isinstance(self.covar, float):
            _noise = self.par['noise']
        if _noise is None:
            raise ValueError('Could not construct noise measurements for Voronoi binning.')
        if _noise.size != self.par['signal'].size:
            raise ValueError('Dimensionality of noise does not match signal.')

        # Down-select to the valid spaxels
        revert_covar = False
        if gpm is None:
            _x = x
            _y = y
            _signal = self.par['signal']
        else:
            _x = x[gpm]
            _y = y[gpm]
            _signal = self.par['signal'][gpm]
            _noise = _noise[gpm]
            if self.covar is not None and isinstance(self.covar, numpy.ndarray):
                revert_covar = True
                _covar = self.covar.copy()
                self.covar = self.covar[gpm,:][:,gpm]

        # All spaxels have S/N greater than threshold, so return each
        # spaxel in its own "bin"
        if numpy.min(_signal/_noise) > self.par['target_snr']:
            warnings.warn('All pixels have enough S/N. Binning is not needed')
            if gpm is None:
                return numpy.arange(x.size, dtype=int)
            binid = numpy.full(x.size, -1, dtype=int)
            binid[gpm] = numpy.arange(numpy.sum(gpm))
            return binid

        # Cannot reach the S/N using all spaxels, so return all spaxels
        # in a single bin
        sn_total = sn_func(numpy.arange(_signal.size), _signal, _noise)
        if sn_total < self.par['target_snr']:
            warnings.warn('Cannot reach target S/N using all data.')
            if gpm is None:
                return numpy.zeros(x.size, dtype=int)
            binid = numpy.full(x.size, -1, dtype=int)
            binid[gpm] = numpy.zeros(numpy.sum(gpm))
            return binid

        # Call the contributed code and return the bin index
        try:
            _binid, xNode, yNode, xBar, yBar, sn, area, scale = \
                voronoi_2d_binning(_x, _y, _signal, _noise, self.par['target_snr'],
                                   sn_func=sn_func, plot=False) #True, quiet=False)
        except:
            warnings.warn('Binning algorithm has raised an exception.  Assume this is because '
                          'all the spaxels should be in the same bin.')
            _binid = numpy.zeros(_signal.size)
        if gpm is None:
            return _binid
        binid = numpy.full(x.size, -1, dtype=int)
        binid[gpm] = _binid
        return binid
# ----------------------------------------------------------------------


# SQUARE BINNING -------------------------------------------------------
# Written by Kate Rubin 9/27/18
# Based on KBW's test_new_binning_scheme.py and the RadialBinning class
class SquareBinningPar(KeywordParSet):
    """
    Class with parameters used by the square binning algorithm.  See
    :class:`mangadap.par.parset.ParSet` for attributes.

    The defined parameters are:

    .. include:: ../tables/squarebinningpar.rst

    """
    def __init__(self, binsz=2.0):
        pars = ['binsz']
        values = [binsz]
        dtypes = [float]
        descr = ['Desired bin size in arcsec']

        super().__init__(pars, values=values, dtypes=dtypes, descr=descr)

    def toheader(self, hdr):
        """
        Copy some of the parameters to a header.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to write to.

        Returns:
            `astropy.io.fits.Header`_: Edited header object

        Raises:
            TypeError:
                Raised if input is not an `astropy.io.fits.Header`_
                object.
        """
        if not isinstance(hdr, fits.Header):
            raise TypeError('Input is not a astropy.io.fits.Header object!')
        hdr['BINSIZE'] = (self['binsz'], 'Square bin size (arcsec)')
        return hdr

    def fromheader(self, hdr):
        """
        Copy the information from the header.

        Args:
            hdr (`astropy.io.fits.Header`_):
                Header object to write to.
        """
        if not isinstance(hdr, fits.Header):
            raise TypeError('Input is not a astropy.io.fits.Header object!')

        self['binsz'] = hdr['BINSIZE']


class SquareBinning(SpatialBinning):
    """
    Class to perform binning of full cube in square apertures
    Length of aperture side is given in arcsec with binsz
    """
    def __init__(self, par=None):
        SpatialBinning.__init__(self, 'square', par=par)
        self.binsz = None

    def bin_index(self, x, y, gpm=None, par=None):
        """
        Bin the data and return the indices of the bins.

        Args:
            x (`numpy.ndarray`_):
                Fiducial on-sky X position of each spectrum. X
                increases with RA.
            y (`numpy.ndarray`_):
                Fiducial on-sky Y position of each spectrum. Y
                increases with DEC.
            gpm (`numpy.ndarray`_, optional):
                Good value (pixel) mask used to select data that should be
                included in the bin.  If None, all values are considered valid.
                If provided, masked values (gpm=False) are given a bin index of
                -1.
            par (:class:`SquareBinningPar`, optional):
                Binning parameters.  Cannot be None.

        Returns:
            `numpy.ndarray`_: An integer bin index for each spectrum.

        Raises:
            ValueError:
                Raised if the sizes of ``x`` and ``y`` do not match or
                if ``par`` is None.
        """
        _x = numpy.asarray(x)
        if len(_x.shape) != 1:
            raise ValueError('On-sky coordinates must be one-dimensional.')
        if len(y) != _x.size:
            raise ValueError('Input coordinates are of different lengths.')
        _y = numpy.asarray(y)

        binid = numpy.full(_x.size, -1, dtype=int)
        if gpm is not None:
            if gpm.size != _x.size:
                raise ValueError(f'Provided GPM has incorrect shape; found {gpm.size}, ' \
                                 f'expected {x.size}.')
            _x = _x[gpm]
            _y = _y[gpm]

        # Perform any necessary initialization
        if par is not None:
            self.par = par
            if self.binsz is not None:
                del self.binsz
                self.binsz = None
        if self.par is None:
            raise ValueError('Required parameters for SquareBinning have not been defined.')
        if self.binsz is None:
            self.binsz = self.par['binsz']



        _binid = numpy.full(_x.size, -1, dtype=int)
        binsz = self.binsz
        minx = numpy.min(_x)-0.25
        maxx = numpy.max(_x)+0.25
        miny = numpy.min(_y)-0.25
        maxy = numpy.max(_y)+0.25

        ctbin = 0

        # Start from the center, work out to edge of each quadrant

        # Quadrant 1
        nslicex = numpy.ceil((maxx) / binsz)
        nslicey = numpy.ceil((maxy) / binsz)

        x_lo = numpy.linspace(0.0,(nslicex*binsz),nslicex+1.0)
        y_lo = numpy.linspace(0.0,(nslicey*binsz),nslicey+1.0)


        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x >= xi) & (_x < xi+binsz) & (_y >= yj) & (_y < yj+binsz)
                if any(indx):
                    _binid[indx] = ctbin
                    ctbin = ctbin + 1


        # Quadrant 2
        nslicex = numpy.ceil((maxx) / binsz)
        nslicey = numpy.floor((miny) / binsz)

        x_lo = numpy.linspace(0.0, (nslicex * binsz), nslicex + 1.0)
        y_lo = numpy.linspace(0.0, (nslicey * binsz), numpy.abs(nslicey) + 1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x >= xi) & (_x < xi + binsz) & (_y <= yj) & (_y > yj - binsz)
                if any(indx):
                    _binid[indx] = ctbin
                    ctbin = ctbin + 1

        # Quadrant 3
        nslicex = numpy.floor((minx) / binsz)
        nslicey = numpy.ceil((maxy) / binsz)

        x_lo = numpy.linspace(0.0, (nslicex * binsz), numpy.abs(nslicex) + 1.0)
        y_lo = numpy.linspace(0.0, (nslicey * binsz), numpy.abs(nslicey) + 1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x <= xi) & (_x > xi - binsz) & (_y >= yj) & (_y < yj + binsz)
                if any(indx):
                    _binid[indx] = ctbin
                    ctbin = ctbin + 1

        # Quadrant 4
        nslicex = numpy.floor((minx) / binsz)
        nslicey = numpy.floor((miny) / binsz)

        x_lo = numpy.linspace(0.0, (nslicex * binsz), numpy.abs(nslicex) + 1.0)
        y_lo = numpy.linspace(0.0, (nslicey * binsz), numpy.abs(nslicey) + 1.0)

        # Find which spaxels land in each aperture
        for xi in x_lo:
            for yj in y_lo:
                indx = (_x <= xi) & (_x > xi - binsz) & (_y <= yj) & (_y > yj - binsz)
                if any(indx):
                    _binid[indx] = ctbin
                    ctbin = ctbin + 1

        if gpm is None:
            return _binid
        binid[gpm] = _binid
        return binid





#!/usr/bin/env python3

import os
import time
import numpy
import glob

from astropy.io import fits
from matplotlib import pyplot, rc, cm, colors, ticker

os.environ['MANGADRP_VER'] = 'v2_4_3'
os.environ['MANGA_SPECTRO_REDUX'] = '/Volumes/MaNGA/redux'
os.environ['MANGADAP_VER'] = '2.2.1'
os.environ['MANGA_SPECTRO_ANALYSIS'] = '/Volumes/MaNGA/analysis'

from mangadap.drpfits import DRPFits

#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.perf_counter()

    font = { 'size' : 16 }
    rc('font', **font)

    cmap = cm.get_cmap('magma')

    ifulist = [  1901,  1902,  3701,  3702,  3703, 3704, 6101, 6102, 6103, 6104, 9101, 9102,
                12701, 12702, 12703, 12704, 12705 ]

#    c = [ 'C0','C0','C1','C1','C1','C1','C2','C2','C2','C2','C3','C3','C4','C4','C4','C4','C4']

    normed_color = colors.Normalize(vmin=0.5, vmax=6.0)
    c = [ cmap(normed_color(1)) ]*2 + [ cmap(normed_color(2)) ]*4 + [ cmap(normed_color(3)) ]*4 \
            + [ cmap(normed_color(4)) ]*2 + [ cmap(normed_color(5)) ]*5

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = fig.add_axes([0.12,0.1,0.8,0.6])
    ax.minorticks_on()
    ax.tick_params(which='major', length=6)
    ax.tick_params(which='minor', length=3)
#    ax.tick_params(which='both', top=True, right=True)
    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-', lw=0.5)

#    ylim = [0.11, 0.59]
    ylim = [0.5, 1.02]
    ax.set_xlim([-0.01, 1.01])
    ax.set_ylim(ylim)

#    for ifu,clr in zip([ifulist[1],ifulist[-2]],[c[1],c[-2]]):
    for ifu,clr in zip(ifulist,c):
        drpf = DRPFits(7495, ifu, 'CUBE', drpver='v2_4_3', read=True)
        print(drpf.file_path())
        flux = drpf.copy_to_masked_array(flag=['DONOTUSE', 'FORESTAR'])
        fgoodpix = numpy.sum(numpy.invert(numpy.ma.getmaskarray(flux)),axis=1) / flux.shape[1]

        if ifu == 12704:
            fgoodpix_saved = fgoodpix.copy().reshape(drpf.spatial_shape)

#        print(fgoodpix.shape)
        srt = numpy.argsort(fgoodpix)
        _fgoodpix = numpy.ma.MaskedArray(fgoodpix[srt], mask=numpy.invert(fgoodpix[srt] > 0))
        omega = 1-(numpy.arange(len(srt), dtype=float)+1)/len(srt)

        omega /= numpy.amax(omega[numpy.invert(_fgoodpix.mask)])

#        print(fgoodpix[srt])
#        ax.step(_fgoodpix, 1-(numpy.arange(len(srt), dtype=float)+1)/len(srt),
#                    color=clr, lw=1.5, zorder=3)
        ax.step(_fgoodpix, omega, color=clr, lw=1.5, zorder=3)

    ax.text(0.50, -0.1, r'$\delta\Lambda$', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.12, 0.50, r'$\delta\Omega (>\delta\Lambda)$', horizontalalignment='center',
            verticalalignment='center', transform=ax.transAxes, rotation='vertical')

    ax.text(0.05, 0.32, r'7495-19xx', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes, color=cmap(normed_color(1)))
    ax.text(0.05, 0.26, r'7495-37xx', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes, color=cmap(normed_color(2)))
    ax.text(0.05, 0.20, r'7495-61xx', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes, color=cmap(normed_color(3)))
    ax.text(0.05, 0.14, r'7495-91xx', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes, color=cmap(normed_color(4)))
    ax.text(0.05, 0.08, r'7495-127xx', horizontalalignment='left', verticalalignment='center',
            transform=ax.transAxes, color=cmap(normed_color(5)))

    ax.plot([0.8,0.8], ylim, dashes=[10,5], lw=2, color='k', zorder=2)

    lim = [-5, fgoodpix_saved.shape[0]+4]

    ax = fig.add_axes([0.45,0.13,0.22,0.22], facecolor='0.9')
    ax.set_xlim(lim)
    ax.set_ylim(lim)
    ax.tick_params(which='both', bottom=False, left=False)
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    cax = fig.add_axes([0.675, 0.13, 0.01, 0.22])
    img = ax.imshow(fgoodpix_saved.T, origin='lower', interpolation='nearest', cmap='gray',
                    vmin=0.5, vmax=None) #0.9)#,
                    #norm=colors.LogNorm(vmin=0.1, vmax=1.0))
    cb = pyplot.colorbar(img, cax=cax)
#    cb.locator = ticker.LogLocator(base=10, subs=(1.,2.,4.,))
#    cb.update_ticks()
    cax.tick_params(labelsize=10)
    cax.text(3.5, 0.95, r'$\delta\Lambda$', horizontalalignment='center',
             verticalalignment='center', transform=cax.transAxes, fontsize=12)

    ax.text(-0.06, 0.05, r'7495-12704', horizontalalignment='center', verticalalignment='bottom',
            transform=ax.transAxes, rotation='vertical')
    
    ofile='../ms/rev1/good_spaxel_growth_v2.pdf'
#    ofile=None
    if ofile is None:
        pyplot.show()
    else:
        fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)


    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






#!/usr/bin/env python3

import os
import time
import numpy
import glob

from IPython import embed

from astropy.io import fits
import astropy.constants

#import matplotlib
#matplotlib.use('Agg')
#from matplotlib import pyplot

from mangadap.dapfits import DAPQualityBitMask, DAPMapsBitMask
from mangadap.config import defaults
from mangadap.util.datatable import DataTable
from mangadap.util.fileio import channel_dictionary


class PPXFFlagTable(DataTable):
    def __init__(self, shape=None):
        datamodel = dict(PLT=dict(typ=int, shape=None),
                         IFU=dict(typ=int, shape=None),
                         CRIT=dict(typ=bool, shape=None),
                         MAIN=dict(typ=bool, shape=None),
                         HAMEAN=dict(typ=float, shape=None),
                         NSPX=dict(typ=int, shape=None),
                         NGOOD=dict(typ=int, shape=None),
                         NFLG=dict(typ=int, shape=None),
                         FLGSNR=dict(typ=float, shape=None),
                         EXSNR=dict(typ=float, shape=None),
                         EXBIN=dict(typ=int, shape=None))
        keys = list(datamodel.keys())
        super(PPXFFlagTable, self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                                            element_shapes=[datamodel[k]['shape'] for k in keys],
                                            shape=shape)


def get_maps_files(dapver='3.0.1', daptype='SPX-MILESHC-MASTARHC2'):
#    return ['/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12704/manga-7443-12704-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12701/manga-7443-12701-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12703/manga-7443-12703-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12705/manga-7443-12705-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/1901/manga-7443-1901-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/1902/manga-7443-1902-MAPS-{1}.fits.gz'.format(dapver, daptype), '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/{0}/{1}/7443/12702/manga-7443-12702-MAPS-{1}.fits.gz'.format(dapver, daptype)]

    root = os.path.join(defaults.dap_analysis_path(dapver=dapver), daptype)
    return glob.glob(os.path.join(root, '*', '*', '*MAPS*fits.gz'))

# This won't work with hybrid binning scheme!
def ppxf_flag_info(tbl, maps_file):
    qual_bm = DAPQualityBitMask()
    bm = DAPMapsBitMask()
   
#    # Targeting bits
#    sdssMaskbits = defaults.sdss_maskbits_file()
#    mngtarg1_bm = BitMask.from_par_file(sdssMaskbits, 'MANGA_TARGET1')
#    mngtarg3_bm = BitMask.from_par_file(sdssMaskbits, 'MANGA_TARGET3')

    tbl['PLT'], tbl['IFU'] = map(lambda x : int(x), os.path.split(maps_file)[1].split('-')[1:3])

    with fits.open(maps_file) as hdu:

        # Check if the file was CRITICAL
        tbl['CRIT'] = qual_bm.flagged(hdu['PRIMARY'].header['DAPQUAL'], flag='DRPCRIT')
        tbl['MAIN'] = hdu[0].header['MNGTARG1'] > 0

        el = channel_dictionary(hdu, 'EMLINE_GFLUX')

        haflx = numpy.ma.MaskedArray(hdu['EMLINE_GFLUX'].data[el['Ha-6564']],
                                     mask=hdu['EMLINE_GFLUX_MASK'].data[el['Ha-6564']] > 0)
        tbl['HAMEAN'] = numpy.ma.median(haflx)
        tbl['NSPX'] = haflx.size
        tbl['NGOOD'] = numpy.sum(numpy.logical_not(numpy.ma.getmaskarray(haflx)))
        indx = bm.flagged(hdu['EMLINE_GFLUX_MASK'].data[el['Ha-6564']], 'FITFAILED')
        tbl['NFLG'] = numpy.sum(indx)

        if not numpy.any(indx):
            return

        snr = numpy.ma.MaskedArray(hdu['SPX_SNR'].data,
                                   mask=numpy.logical_not(hdu['SPX_SNR'].data > 0))
        tbl['FLGSNR'] = numpy.ma.median(snr[indx])
        srt = numpy.argsort(snr[indx])
        tbl['EXSNR'] = snr[indx][srt][-1]
        tbl['EXBIN'] = hdu['BINID'].data[3][indx][srt][-1]


def main():
    ofile = 'ppxf_failures_hyb.fits'
    force = False
    if os.path.isfile(ofile) and not force:
        print('File already exists.')
        return

    daptype = 'HYB10-MILESHC-MASTARHC2'
    maps_files = get_maps_files()
    nmaps = len(maps_files)
    print('Found {0} MAPS files.'.format(nmaps))

    tbl = PPXFFlagTable(shape=nmaps)

    # Iterate through the maps files and add the data
    for i in range(nmaps):
        print('{0}/{1}'.format(i+1,nmaps), end='\r')
        ppxf_flag_info(tbl[i], maps_files[i])
    print('{0}/{0}'.format(nmaps))

    # Write the output
    print('Writing: {0}'.format(ofile))
    fits.HDUList([fits.PrimaryHDU(), tbl.to_hdu(name='PAR')]).writeto(ofile, overwrite=True)

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.perf_counter()
    main()
    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))






# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of utility functions to deal with dust extinction.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings
import numpy
from numpy.polynomial.polynomial import polyval
import scipy.interpolate

def default_calzetti_rv():
    return 4.05

def reddening_vector_calzetti(wave, ebv, rv=None):
    r"""
    Return the Calzetti et al. (2000) reddening vector.

    Args:
        wave (array-like): Wavelengths at which to calculate the
            reddening vector curve in angstroms.
        ebv (float): E(B-V) reddening used to normalize the curve.
        rv (float): (**Optional**) Ratio of V-band extinction to the B-V
            reddening:

            .. math:: 

                R_V = \frac{A_V}{E(B-V)}

            Default is 4.05.  Typical value for the diffuse ISM of the
            Milky Way is 3.1.

    Returns:
        numpy.ma.MaskedArray: One-dimensional masked array with the
        reddening vector that can be used deredden a spectrum by
        calculating:

        .. math::

            F(\lambda) = a(\lambda) f(\lambda)

        where :math:`a` is the vector returned by this function,
        :math:`f` is the observed flux, and :math:`F` is the dereddened
        flux.
    """
    # Check shapes
    _wave = numpy.atleast_1d(wave)
    if len(_wave.shape) != 1:
        raise ValueError('Must only provide a single wavlength vector.')
    if not isinstance(ebv, float):
        raise TypeError('Input reddening value must be a single float.')

    # Wavenumber in 1/micron
    k = 1e4/_wave

    _rv = default_calzetti_rv() if rv is None else rv

    # Select valid wavelength ranges
    w1 = (_wave > 6300.) & (_wave < 22000.)
    w2 = (_wave > 912.) & (_wave <= 6300.)
    if numpy.sum(w1) + numpy.sum(w2) != _wave.size:
        warnings.warn('Invalid wavelength range in input vector.  Invalid regions will be masked.')

    # Extinction curve
    ext = numpy.ma.zeros(_wave.size, dtype=float)
    ext[numpy.invert(w1) & numpy.invert(w2)] = numpy.ma.masked
    ext[w1] = 2.659*(-1.857 + 1.040*k[w1]) + _rv
    ext[w2] = 2.659*(polyval(k[w2], [-2.156, 1.509, -0.198, 0.011])) + _rv

    # Return dereddening vector
    return numpy.ma.power(10., 0.4*ext*ebv)


def default_ccm_rv():
    return 3.1

def reddening_vector_ccm(wave, ebv, rv=None, original=False):
    r"""

    Return the reddening vector based on Cardelli, Clayton, and Mathis
    (1989 ApJ.  345, 245), including the update for the near-UV given by
    O'Donnell (1994, ApJ, 422, 158).   Parameterization is valid from
    the IR to the far-UV (3.5 microns to 0.1 microns). 

    Args:
        wave (array-like): Wavelengths at which to calculate the
            reddening vector in angstroms.
        ebv (float): E(B-V) reddening used to normalize the curve.
        rv (float): (**Optional**) Ratio of V-band extinction to the B-V
            reddening:

            .. math:: 

                R_V = \frac{A_V}{E(B-V)}

            Default is the typical value for the diffuse ISM of the
            Milky Way, :math:`R_V = 3.1`.
        original (bool): (**Optional**) Use the original coefficients
            from CCM89 instead of the updated coefficients from
            O'Donnell (1994).  Default is to use the updated
            coefficients.

    Returns:
        numpy.ma.MaskedArray: One-dimensional masked array with the
        reddening vector that can be used deredden a spectrum by
        calculating:

        .. math::

            F(\lambda) = a(\lambda) f(\lambda)

        where :math:`a` is the vector returned by this function,
        :math:`f` is the observed flux, and :math:`F` is the dereddened
        flux.
    """
    # Check shapes
    _wave = numpy.atleast_1d(wave)
    if len(_wave.shape) != 1:
        raise ValueError('Must only provide a single wavlength vector.')
    if not isinstance(ebv, float):
        raise TypeError('Input reddening value must be a single float.')

    # Wavenumber in 1/micron
    k = 1e4/_wave

    _rv = default_ccm_rv() if rv is None else rv

    a = numpy.zeros(k.size, dtype=float)
    b = numpy.zeros(k.size, dtype=float)

    # Compute the Infrared portion
    w1 = (k > 0.3) & (k < 1.1)
    a[w1] =  0.574 * numpy.power(k[w1],1.61)
    b[w1] = -0.527 * numpy.power(k[w1],1.61)

    # Compute the Optical/NIR portion
    if original:
        c1 = numpy.array([      1.,  0.17699, -0.50447, -0.02427,  0.72085,  0.01979, -0.77530,
                           0.32999 ])
        c2 = numpy.array([      0.,  1.41338,  2.28305,  1.07233, -5.38434, -0.62251,  5.30260,
                          -2.09002 ])
    else:
        c1 = numpy.array([     1., 0.104,   -0.609,    0.701,  1.137, -1.718,   -0.827,    1.647,
                           -0.505 ])
        c2 = numpy.array([     0., 1.952,    2.908,   -3.989, -7.985, 11.102,    5.491,  -10.805,
                            3.347 ])
    w1 = (k >= 1.1) & (k < 3.3)
    a[w1] = polyval(k[w1] - 1.82, c1)
    b[w1] = polyval(k[w1] - 1.82, c2)

    # Compute the mid-UV portion
    w1 = (k >= 3.3) & (k < 8.)

    w2 = w1 & (k > 5.9)
    fa = numpy.zeros(k.size, dtype=float)
    fb = numpy.zeros(k.size, dtype=float)
    fa[w2] = -0.04473 * numpy.square(k[w2]-5.9) - 0.009779 * numpy.power(k[w2]-5.9,3)
    fb[w2] =   0.2130 * numpy.square(k[w2]-5.9) +   0.1207 * numpy.power(k[w2]-5.9,3)

    a[w1] =  1.752 - 0.316*k[w1] - (0.104 / ( numpy.square(k[w1]-4.67) + 0.341 )) + fa[w1]
    b[w1] = -3.090 + 1.825*k[w1] + (1.206 / ( numpy.square(k[w1]-4.62) + 0.263 )) + fb[w1]

    # Compute the far-UV portion
    w1 = (k >= 8.) & (k <= 11.)
    a[w1] = polyval(k[w1]-8., [ -1.073, -0.628,  0.137, -0.070 ])
    b[w1] = polyval(k[w1]-8., [ 13.670,  4.257, -0.420,  0.374 ])

    ext = numpy.ma.MaskedArray(_rv*(a+b/_rv))
    ext[ (k<=0.3) | (k>11.) ] = numpy.ma.masked

    # Return dereddening vector
    return numpy.ma.power(10., 0.4*ext*ebv)


class FMExtinctionCoefficients:
    def __init__(self, k0, gamma, c1, c2, c3, c4):
        self.k0 = k0
        self.gamma = gamma
        self.c1 = c1
        self.c2 = c2
        self.c3 = c3
        self.c4 = c4

    @classmethod
    def from_Rv(cls, rv):
        c2 = -0.824 + 4.717/rv
        return cls(4.596, 0.99, 2.030 - 3.007*c2, c2, 3.23, 0.41)


class AvgLMCExtinctionCoefficients(FMExtinctionCoefficients):
    def __init__(self):
        FMExtinctionCoefficients.__init__(self, 4.596, 0.91, -1.28, 1.11, 2.73, 0.64)


class LMC2ExtinctionCoefficients(FMExtinctionCoefficients):
    def __init__(self):
        FMExtinctionCoefficients.__init__(self, 4.626, 1.05, -2.16, 1.31, 1.92, 0.42)


def default_fm_rv():
    return 3.1


def reddening_vector_fm(wave, ebv, rv=None, coeffs=None):
    r"""

    Return the reddening vector based on Fitzpatrick & Massa
    (Fitzpatrick, 1999, PASP, 111, 63; astro-ph/9809387 ).
    Parameterization is valid from the IR to the far-UV (3.5 microns to
    0.1 microns). UV extinction curve is extrapolated down to 912
    Angstroms.

    Args:
        wave (array-like): Wavelengths at which to calculate the
            reddening vector in angstroms.
        ebv (float): E(B-V) reddening used to normalize the curve.
        rv (float): (**Optional**) Ratio of V-band extinction to the B-V
            reddening:

            .. math:: 

                R_V = \frac{A_V}{E(B-V)}

            Default is the typical value for the diffuse ISM of the
            Milky Way, :math:`R_V = 3.1`.
        coeffs (:class:`FMExtinctionCoefficients`): (**Optional**)
            Object with the coefficients to use for the extinction
            curve.  Default is to use the :math:`R_V` dependent
            coefficients defined using the
            :func:`FMExtinctionCoefficients.from_Rv` class method.

    Returns:
        numpy.ma.MaskedArray: One-dimensional masked array with the
        reddening vector that can be used deredden a spectrum by
        calculating:

        .. math::

            F(\lambda) = a(\lambda) f(\lambda)

        where :math:`a` is the vector returned by this function,
        :math:`f` is the observed flux, and :math:`F` is the dereddened
        flux.
    """
    # Check shapes
    _wave = numpy.atleast_1d(wave)
    if len(_wave.shape) != 1:
        raise ValueError('Must only provide a single wavlength vector.')
    if not isinstance(ebv, float):
        raise TypeError('Input reddening value must be a single float.')

    # Check the provided coefficients
    if coeffs is not None and not isinstance(coeffs, FMExtinctionCoefficients):
        raise TypeError('Coefficients must be provided by a FMExtinctionCoefficients object.')

    _rv = default_fm_rv() if rv is None else rv
    _coeffs = FMExtinctionCoefficients.from_Rv(_rv) if coeffs is None else coeffs

    # Wavenumber in 1/micron
    k = 1e4/_wave
    ext = numpy.ma.zeros(_wave.size, dtype=float)
    
    # UV portion
    w1 = k > 1e4/2700.
    splpts_uv_k = 1e4/numpy.array([2700.,2600.])
    uv_k = numpy.append(splpts_uv_k, k[w1]) if numpy.sum(w1) > 0 else splpts_uv_k

    uv_k2 = numpy.square(uv_k)
    uv_kclip = (uv_k-5.9).clip(0.,None)
    uv_ext = _coeffs.c1  + _coeffs.c2*uv_k \
             + _coeffs.c3*uv_k2/(numpy.square(uv_k2 - _coeffs.k0**2) + uv_k2*_coeffs.gamma**2) \
             + _coeffs.c4*(0.5392*numpy.square(uv_kclip)+0.05644*numpy.power(uv_kclip,3)) + _rv

    # Pull out spline points
    splpts_uv_ext, ext[w1] = uv_ext[:2], uv_ext[2:] 

    # Return if no optical/IR part
    if numpy.sum(numpy.invert(w1)) == 0:
        return numpy.ma.power(10., 0.4*ext*ebv)

    # Optical/IR portion
    splpts_oi_k = numpy.append([0],10000.0/numpy.array([26500.0, 12200.0, 6000.0, 5470.0, 4670.0,
                                                        4110.0]))

    splpts_oi_ext = numpy.append( numpy.array([0.0,0.26469,0.82925])*_rv/3.1,
                                  numpy.array([ polyval(_rv, [-4.22809e-01, 1.00270,  2.13572e-04]),
                                                polyval(_rv, [-5.13540e-02, 1.00216, -7.35778e-05]),
                                                polyval(_rv, [ 7.00127e-01, 1.00184, -3.32598e-05]),
                                                polyval(_rv, [     1.19456, 1.01707, -5.46959e-03,
                                                               7.97809e-04, -4.45636e-05]) ]) )
    
    # Use bc_type = 'natural' to be identical to fm_unred.pro in
    # IDLUTILS
    interp = scipy.interpolate.CubicSpline(numpy.append(splpts_oi_k,splpts_uv_k),
                                           numpy.append(splpts_oi_ext,splpts_uv_ext),
                                           bc_type='natural')
    w2 = numpy.invert(w1)
    ext[w2] = interp(k[w2])
    return numpy.ma.power(10., 0.4*ext*ebv)


class GalacticExtinction:
    r"""
    Class that uses the above functions to compute and apply the
    Galactic reddening curve according to the desired form.

    Args:
        form (str): (**Optional**) The form of the reddening curve to
            adopt.  The default is 'ODonnell'

        wave (numpy.ndarray): (**Optional**) The wavelengths for the
            computation of the reddening vector to compute.  Default is
            to leave the vector undefined.

        ebv (float): (**Optional**) E(B-V) reddening used to normalize
            the curve.  Default is to leave the vector undefined.

        rv (float): (**Optional**) Ratio of V-band extinction to the B-V
            reddening:

            .. math:: 

                R_V = \frac{A_V}{E(B-V)}

            Default value depends on the requested form.  For 'FM',
            'CCM', and 'ODonnell', the default is 3.1 (typical for the
            diffuse ISM of the Milky Way); for 'Calzetti', it is 4.05.
    
        coeffs (:class:`FMExtinctionCoefficients`): (**Optional**)
            Object with the coefficients to use for the extinction
            curve if the form is 'FM'.  If the form is **not** 'FM',
            these coefficients are ignored (and should be provided as
            None).  If the form is 'FM' and `coeffs=None`, the default
            is to use the :math:`R_V` dependent coefficients defined
            using the :func:`FMExtinctionCoefficients.from_Rv` class
            method.

    Raises:
        ValueError: Raised if *form* is not a known form.

    """
    def __init__(self, form='ODonnell', wave=None, ebv=None, rv=None, coeffs=None):
        if form not in GalacticExtinction.valid_forms():
            raise ValueError('Unrecognized form of the extinction law: {0}'.format(form))
        self.form = form

        self.wave = None
        self.ebv = None
        self.rv = None
        self.coeffs = None
        self.redcorr = None

        self.compute(wave, ebv, rv=rv, coeffs=coeffs)


    @staticmethod
    def valid_forms():
        """
        Returns a list keyword for the allowed reddening curves forms
        implemented within the :class:`GalacticExtinction` class.
        """
        return [ None, 'ODonnell', 'CCM', 'FM', 'Calzetti']

    
    def _deredden_vec(self):
        return self.redcorr


    def _redden_vec(self):
        return numpy.ma.power(self.redcorr, -1)


    def compute(self, wave, ebv, rv=None, coeffs=None):
        r"""
        Compute the reddening curve of the desired form.

        Args:
            wave (numpy.ndarray): The wavelengths for the computation of
                the reddening correction to compute.

            ebv (float): E(B-V) reddening used to normalize the curve.

            rv (float): (**Optional**) Ratio of V-band extinction to the
                B-V reddening:

                .. math:: 

                    R_V = \frac{A_V}{E(B-V)}

                Default value depends on the requested form.  For 'FM',
                'CCM', and 'ODonnell', the default is 3.1 (typical for
                the diffuse ISM of the Milky Way); for 'Calzetti', it is
                4.05.
    
            coeffs (:class:`FMExtinctionCoefficients`): (**Optional**)
                Object with the coefficients to use for the extinction
                curve if the form is 'FM'.  If the form is **not** 'FM',
                these coefficients are ignored (and should be provided
                as None).  If the form is 'FM' and `coeffs=None`, the
                default is to use the :math:`R_V` dependent coefficients
                defined using the
                :func:`FMExtinctionCoefficients.from_Rv` class method.

        Returns:
            numpy.ndarray : Vector with the extinction at every provided
            wavelength.

        """
        self.wave = wave
        self.ebv = ebv
        if rv is None:
            if self.form in [ 'ODonnell', 'CCM' ]:
                self.rv = default_ccm_rv()
            elif self.form == 'FM':
                self.rv = default_fm_rv()
            elif self.form == 'Calzetti':
                self.rv = default_calzetti_rv()
        else:
            self.rv = rv

        self.coeffs = FMExtinctionCoefficients.from_Rv(self.rv) \
                            if coeffs is None and self.form == 'FM' else coeffs

        if self.wave is None or self.ebv is None:
            self.redcorr = None
            return self.redcorr

        if self.form == 'ODonnell':
            self.redcorr = reddening_vector_ccm(self.wave, self.ebv, rv=self.rv, original=False)
        elif self.form == 'CCM':
            self.redcorr = reddening_vector_ccm(self.wave, self.ebv, rv=self.rv, original=True)
        elif self.form == 'FM':
            self.redcorr = reddening_vector_fm(self.wave, self.ebv, rv=self.rv, coeffs=self.coeffs)
        elif self.form == 'Calzetti':
            self.redcorr = reddening_vector_calzetti(self.wave, self.ebv, rv=self.rv)
        else:
            self.redcorr = numpy.ones(self.wave.size, dtype=float)

        return self.redcorr


    def apply(self, flux, ivar=None, err=None, deredden=True):
        r"""
        Redden or deredden the provided flux array using the existing
        reddening correction vector.

        If flux has a dimensionality larger than 1, it is assumed that
        the spectra are organized along the last axis and the reddening
        correction is applied to all spectra in the array.

        If ivar is provided, the error is propagated.

        Args:
            flux (numpy.ndarray): The flux array to correct.
            ivar (numpy.ndarray): (**Optional**) The flux inverse
                variance to use for the error propagation.  If provided,
                inverse variance in the flux array will be returned.  If
                provided, cannot provide 1-sigma errors.
            err (numpy.ndarray): (**Optional**) The 1-siga errors in the
                flux to use for the error propagation.  If provided,
                errors in the flux array will be returned.  If provided,
                cannot provide inverse variance.
            deredden (bool): (**Optional**) Flag to **deredden** the
                spectrum; if set to false, the function will instead
                redden the provided fluxes.

        Returns:
            numpy.ndarray : Returns the reddened flux array, and the
            inverse variance if ivar is provided.

        Raises:
            ValueError: Raised if the internal reddening correction
                vector is not defined, if the entered flux and ivar
                arrays do not have the same shape, or if the number of
                wavelength channels in flux does not match the expected
                number based on the length of :attr:`redcorr`.
        """
        # Check the input
        if self.redcorr is None:
            raise ValueError('Must first calculate reddening correction vector.')
        if ivar is not None and flux.shape != ivar.shape:
            raise ValueError('Flux and inverse variance arrays must have the same shape.')
        if ivar is not None and err is not None:
            raise ValueError('Cannot provide both errors and inverse variance.')
        if err is not None and flux.shape != err.shape:
            raise ValueError('Flux and error arrays must have the same shape.')

        r = self._deredden_vec() if deredden else self._redden_vec()

        if flux.shape[-1] != self.redcorr.size:
            raise ValueError('Flux and reddening correction vector must have same size.')
        if ivar is None and err is None:
            return flux * r[...,:]
        elif err is None:
            return flux * r[...,:], ivar / numpy.square(r)[...,:]
        return flux * r[...,:], err * r[...,:]




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""
Defines the class that constructs the DAPall summary table.

This is a post processing script that **must** be run after the DRPall
file is created.

.. todo::

    - Include columns with just the absolute luminosity of the
      H-alpha line as the intermediate step between the apparent
      luminosity and the star-formation rate.
    - Provide a rough Balmer decrement measurement?

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import logging
import time
import os
import warnings
import numpy

from scipy import interpolate

from astropy.io import fits
import astropy.constants
from astropy.cosmology import FlatLambdaCDM
import astropy.units
from astropy.stats import sigma_clip

from .drpcomplete import DRPComplete
from ..dapfits import DAPMapsBitMask, DAPQualityBitMask
from ..util.drpbitmask import DRPQuality3DBitMask
from ..config import defaults
from ..util.fileio import init_record_array, rec_to_fits_type, channel_dictionary, channel_units
from ..util.fitsutil import DAPFitsUtil
from ..util.log import init_DAP_logging, module_logging, log_output
from ..config.analysisplan import AnalysisPlan
from ..par.absorptionindexdb import AbsorptionIndexDB
from ..par.bandheadindexdb import BandheadIndexDB
from ..par.emissionlinedb import EmissionLineDB
from ..par.emissionmomentsdb import EmissionMomentsDB
from ..proc.util import select_proc_method, sample_growth
from ..proc.spatiallybinnedspectra import SpatiallyBinnedSpectra
from ..proc.stellarcontinuummodel import StellarContinuumModel
from ..proc.emissionlinemoments import EmissionLineMomentsDef
#from ..proc.emissionlinemoments import available_emission_line_moment_databases
from ..proc.emissionlinemodel import EmissionLineModel, EmissionLineModelDef
#from ..proc.emissionlinemodel import available_emission_line_modeling_methods
from ..proc.spectralindices import SpectralIndicesDef #, available_spectral_index_databases

from matplotlib import pyplot

class DAPall:
    """
    Construct the summary table information for the DAP.

    Any observation in the
    :class:`mangadap.survey.drpcomplete.DRPComplete` file that
    satisfies the criteria set by
    :func:`mangadap.survey.drpcomplete.DRPComplete.can_analyze` is
    considered. If the expected output files exist, they are analyzed
    and included in the DAPall database. If the expected files do not
    exist, the construction of the DAPall (and the relevant flag)
    assumes that the DAP analysis of this cube failed.

    Args:
        plan (:class:`mangadap.par.analysisplan.AnalysisPlanSet`):
            The plan object used by the DAP.
        methods (:obj:`str`, :obj:`list`, optional):
            Specify a set of methods in the DAP plan file to include
            in the DAPall file. If not provided, all methods in the
            plan file are included. Each method produces an extension
            in the DAPall file, where the order of the extensions is
            alphanumeric.
        drpver (:obj:`str`, optional):
            DRP version.  Default determined by
            :func:`mangadap.config.defaults.drp_version`.
        redux_path (:obj:`str`, optional):
            Top-level directory with the DRP products; default is
            defined by
            :func:`mangadap.config.defaults.drp_redux_path`.
        dapver (:obj:`str`, optional):
            DAP version.  Default determined by
            :func:`mangadap.config.defaults.dap_version`.
        analysis_path (:obj:`str`, optional):
            Top-level directory for the DAP output data; default is
            defined by
            :func:`mangadap.config.defaults.dap_analysis_path`.
        readonly (:obj:`bool`, optional):
            If it exists, open any existing file and disallow any
            modifications of the database. Default is to
            automatically build the entire database if it doesn't
            exist.
        loggers (:obj:`list`, optional):
            List of `logging.Logger`_ objects used to log progress.
        quiet (:obj:`bool`, optional):
            Suppress all terminal and logging output.

    Raises:
        TypeError: Raised if the input plan is not a
            :class:`mangadap.par.analysisplan.AnalysisPlanSet` object.
        FileNotFoundError: Raise if the DRPall or the DRPComplete file
            are not available.
    
    Attributes:
        drpver (:obj:`str`):
            DRP version
        redux_path (:obj:`str`):
            Path to top-level redux directory
        dapver (:obj:`str`):
            DAP version
        analysis_path (:obj:`str`):
            Path to top-level analysis directory
        drpall_file (:obj:`str`):
            Path to the DRPall file
        h (:obj:`float`):
            The Hubble parameter (currently always set to unity)
        H0 (:obj:`float`):
            Hubbles constant (km/s/Mpc)
        cosmo (`astropy.cosmology.FlatLambdaCDM`_):
            Object used for distance calculations.
        maps_bm (:class:`mangadap.dapfits.DAPMapsBitMask`):
            Bitmask used to mask map pixels.
        neml (:obj:`int`):
            Number of emission lines included in each method.
        nindx (:obj:`int`):
            Number of spectral indices included in each method.
        str_len (:obj:`dict`):
            Dictionary with the number of characters used for each
            string entry in the database.
        hdu (`astropy.io.fits.HDUList`_):
            Object with the table data.
        ndap (:obj:`int`):
            Number of rows in the data table; equivalent to the
            number of processed MAPS files.
        plate (`numpy.ndarray`_):
            Array of the plate numbers
        ifudesign (`numpy.ndarray`_):
            Array of the ifudesigns
        methods (`numpy.ndarray`_):
            Array of the methods included in the DAPall file.
        readonly (:obj:`bool`):
            Object is prohibited from updating the database.
        loggers (:obj:`list`):
            List of `logging.Logger`_ objects to log progress;
            ignored if quiet=True.
        quiet (:obj:`bool`):
            Suppress all terminal and logging output.
        float_dtype (:obj:`str`):
            Sets precision for floating-point numbers.
    """
    def __init__(self, plan, methods=None, drpver=None, redux_path=None, dapver=None,
                 analysis_path=None, readonly=False, loggers=None, quiet=False,
                 single_precision=False):

        # Initialize the reporting
        self.loggers = loggers
        self.quiet = quiet

        self.float_dtype = 'float32' if single_precision else 'float'

        # Path definitions
        self.drpver = defaults.drp_version() if drpver is None else str(drpver)
        self.redux_path = defaults.drp_redux_path(self.drpver) \
                                if redux_path is None else str(redux_path)

        self.dapver = defaults.dap_version() if dapver is None else str(dapver)
        self.analysis_path = defaults.dap_analysis_path(self.drpver, self.dapver) \
                                if analysis_path is None else str(analysis_path)

        # Set the name of the DRPall file
        self.drpall_file = os.path.join(self.redux_path, 'drpall-{0}.fits'.format(self.drpver))
        if not os.path.isfile(self.drpall_file):
            raise FileNotFoundError('{0} does not exist!'.format(self.drpall_file))

        # Initialize the bitmask
        self.maps_bm = DAPMapsBitMask()

        # Set the length for the string elements of the table
        self.str_len = self._init_string_lengths()

        # Initialize the cosmology object
        self.h = 1.0
        self.H0 = 100 * self.h * astropy.units.km / astropy.units.s / astropy.units.Mpc
        self.cosmo = FlatLambdaCDM(H0=self.H0, Om0=0.3)

        # Declare the other attributes for use later on
        self.nmom = None
        self.elmom_channels = None
        self.neml = None
        self.elfit_channels = None
        self.nindx = None
        self.spindx_channels = None
        self.spindx_units = None
        self.hdu = None
        self.ndap = None
        self.plate = None
        self.ifudesign = None
        self.methods = None

        # Give an initial report
        self.readonly = readonly
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, '{0:^50}'.format('DAPALL DATABASE'))
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, 'DRPall file: {0}'.format(self.drpall_file))
            log_output(self.loggers, 1, logging.INFO, 'Output path: {0}'.format(self.analysis_path))
            log_output(self.loggers, 1, logging.INFO, 'Output file: {0}'.format(self.file_name()))
            if self.readonly:
                log_output(self.loggers, 1, logging.INFO, 'READ-ONLY MODE')

        # If the output file already exists, read it
        if os.path.isfile(self.file_path()):
            self._read()

        # Check the combination of read-only and the read data makes
        # sense
        if self.readonly and self.hdu is None:
            raise ValueError('Selected read-only, but no database can be read.')
        if not self.readonly and self.hdu is not None:
            warnings.warn('DAPall file already exists!  Update will overwrite existing data.')

        # Automatically attempt to update the database if no database
        # exists or if the database is not meant to be read-only
        if self.hdu is None or not self.readonly:
            self.update(plan, methods=methods)

#    def __getitem__(self, key):
#        return self.hdu['DAPALL'].data[key]

    @staticmethod
    def _emission_line_moment_db_info(key):
        if key == 'None':
            return 'None', None
        db = select_proc_method(key, EmissionLineMomentsDef,
                                available_func=available_emission_line_moment_databases)
        return db['passbands'], EmissionMomentsDB.from_key(db['passbands']).channel_names()

    @staticmethod
    def _emission_line_db_info(key):
        if key == 'None':
            return 'None', None
        db = select_proc_method(key, EmissionLineModelDef,
                                available_func=available_emission_line_modeling_methods)
        return db['emission_lines'], EmissionLineDB.from_key(db['emission_lines']).channel_names()

    @staticmethod
    def _spectral_index_db_info(key):
        if key == 'None':
            return 0, 'None', 'None', None
        db = select_proc_method(key, SpectralIndicesDef,
                                available_func=available_spectral_index_databases)
        absdb = AbsorptionIndexDB.from_key(db['absindex'])
        bhddb = BandheadIndexDB.from_key(db['bandhead'])
        nindx = absdb.size + bhddb.size
        units = numpy.array(absdb['units'].tolist() + ['']*bhddb.size)
        channels = absdb.channel_names()
        channels.update(bhddb.channel_names(offset=absdb.size))
        if len(channels) != nindx:
            raise ValueError('Spectral index channel names not unique!')
        return db['absindex'], db['bandhead'], channels, units

    # TODO: Fix this!
    def _read(self):
        """
        Read the data in the existing file at :func:`file_path`.
        """
        # Close the HDUList if it's already open
        if self.hdu is not None:
            self.hdu.close()
            self.hdu = None

        # Read the file
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Reading exiting file')
        self.hdu = fits.open(self.file_path())

        # Methods are the extension names
        self.methods = [h.name for h in self.hdu][1:]

        # Number of cubes analyzed by the DAP
        self.ndap = self.hdu[self.methods[0]].header['NAXIS2']

        # Find the plate-IFU combinations in the file
        pltifu = numpy.unique(self.hdu[self.methods[0]].data['PLATEIFU'])
        self.plate, self.ifudesign = map(lambda x: numpy.array(x),
                                         numpy.array([ [int(p),int(f)] 
                                                        for p,f in pltifu.split('-')]).T.tolist())

        # Set the number of emission lines and spectral indices
        # TODO: Get this from the number of header keywords?
        self.nmom = self.hdu[self.methods[0]].data['EMLINE_SFLUX_1RE'].shape[1]
        self.neml = self.hdu[self.methods[0]].data['EMLINE_GFLUX_1RE'].shape[1]
        self.nindx = self.hdu[self.methods[0]].data['SPECINDEX_1RE'].shape[1]

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Rows per method: {0}'.format(self.ndap))
            log_output(self.loggers, 1, logging.INFO, 'Methods: {0}'.format(self.methods))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of observations: {0}'.format(len(self.plate)))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission-line moments: {0}'.format(self.nmom))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission lines: {0}'.format(self.neml))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of spectral indices: {0}'.format(self.nindx))

    def _init_string_lengths(self):
        return { 'PLATEIFU':12, 
                 'MANGAID':10,
                 'MODE':4,
                 'DAPTYPE':30,
                 'VERSDRP2':8,
                 'VERSDRP3':8,
                 'VERSCORE':8,
                 'VERSUTIL':8,
                 'VERSDAP':8,
                 'RDXQAKEY':14,
                 'BINKEY':14,
                 'SCKEY':14,
                 'ELMKEY':14,
                 'ELFKEY':14,
                 'SIKEY':14,
                 'BINTYPE':10,
                 'TPLKEY':15,
                 'DATEDAP':10,
                 'EMLMOM_NAME':20,
                 'EMLINE_NAME':20,
                 'SPECINDEX_NAME':20,
                 'SPECINDEX_UNIT':5
               }

    def _table_dtype(self, nmom, neml, nindx):
        return [ ('PLATE', int),
                 ('IFUDESIGN', int),
                 ('PLATEIFU', '<U{0:d}'.format(self.str_len['PLATEIFU'])),
                 ('MANGAID', '<U{0:d}'.format(self.str_len['MANGAID'])),
                 ('DRPALLINDX', int),
                 ('MODE', '<U{0:d}'.format(self.str_len['MODE'])),
                 ('DAPTYPE', '<U{0:d}'.format(self.str_len['DAPTYPE'])),
                 ('DAPDONE', bool),
                 ('OBJRA', self.float_dtype),
                 ('OBJDEC', self.float_dtype),
                 ('IFURA', self.float_dtype),
                 ('IFUDEC', self.float_dtype),
                 ('MNGTARG1', int),
                 ('MNGTARG2', int),
                 ('MNGTARG3', int),
                 ('Z', self.float_dtype),
                 ('LDIST_Z', self.float_dtype),
                 ('ADIST_Z', self.float_dtype),
                 ('NSA_Z', self.float_dtype),
                 ('NSA_ZDIST', self.float_dtype),
                 ('LDIST_NSA_Z', self.float_dtype),
                 ('ADIST_NSA_Z', self.float_dtype),
                 ('NSA_ELPETRO_BA', self.float_dtype),
                 ('NSA_ELPETRO_PHI', self.float_dtype),
                 ('NSA_ELPETRO_TH50_R', self.float_dtype),
                 ('NSA_SERSIC_BA', self.float_dtype),
                 ('NSA_SERSIC_PHI', self.float_dtype),
                 ('NSA_SERSIC_TH50', self.float_dtype),
                 ('NSA_SERSIC_N', self.float_dtype),
                 ('VERSDRP2', '<U{0:d}'.format(self.str_len['VERSDRP2'])),
                 ('VERSDRP3', '<U{0:d}'.format(self.str_len['VERSDRP3'])),
                 ('VERSCORE', '<U{0:d}'.format(self.str_len['VERSCORE'])),
                 ('VERSUTIL', '<U{0:d}'.format(self.str_len['VERSUTIL'])),
                 ('VERSDAP', '<U{0:d}'.format(self.str_len['VERSDAP'])),
                 ('DRP3QUAL', int),
                 ('DAPQUAL', int),
                 ('RDXQAKEY', '<U{0:d}'.format(self.str_len['RDXQAKEY'])),
                 ('BINKEY', '<U{0:d}'.format(self.str_len['BINKEY'])),
                 ('SCKEY', '<U{0:d}'.format(self.str_len['SCKEY'])),
                 ('ELMKEY', '<U{0:d}'.format(self.str_len['ELMKEY'])),
                 ('ELFKEY', '<U{0:d}'.format(self.str_len['ELFKEY'])),
                 ('SIKEY', '<U{0:d}'.format(self.str_len['SIKEY'])),
                 ('BINTYPE', '<U{0:d}'.format(self.str_len['BINTYPE'])),
                 ('BINSNR', self.float_dtype),
                 ('TPLKEY', '<U{0:d}'.format(self.str_len['TPLKEY'])),
                 ('DATEDAP', '<U{0:d}'.format(self.str_len['DATEDAP'])),
                 ('DAPBINS', int),
                 ('RCOV90', self.float_dtype),
                 ('SNR_MED', self.float_dtype, (4,)),
                 ('SNR_RING', self.float_dtype, (4,)),
                 ('SB_1RE', self.float_dtype),
                 ('BIN_RMAX', self.float_dtype),
                 ('BIN_R_N', self.float_dtype, (3,)),
                 ('BIN_R_SNR', self.float_dtype, (3,)),
                 ('STELLAR_Z', self.float_dtype),
                 ('STELLAR_VEL_LO', self.float_dtype),
                 ('STELLAR_VEL_HI', self.float_dtype),
                 ('STELLAR_VEL_LO_CLIP', self.float_dtype),
                 ('STELLAR_VEL_HI_CLIP', self.float_dtype),
                 ('STELLAR_SIGMA_1RE', self.float_dtype),
                 ('STELLAR_RCHI2_1RE', self.float_dtype),
                 ('HA_Z', self.float_dtype),
                 ('HA_GVEL_LO', self.float_dtype),
                 ('HA_GVEL_HI', self.float_dtype),
                 ('HA_GVEL_LO_CLIP', self.float_dtype),
                 ('HA_GVEL_HI_CLIP', self.float_dtype),
                 ('HA_GSIGMA_1RE', self.float_dtype),
                 ('HA_GSIGMA_HI', self.float_dtype),
                 ('HA_GSIGMA_HI_CLIP', self.float_dtype),
                 ('EMLINE_RCHI2_1RE', self.float_dtype),
#                 ('EMLMOM_NAME', '<U{0:d}'.format(self.str_len['EMLMOM_NAME']*nmom)),
                 ('EMLINE_SFLUX_CEN', self.float_dtype, (nmom,)),
                 ('EMLINE_SFLUX_1RE', self.float_dtype, (nmom,)),
                 ('EMLINE_SFLUX_TOT', self.float_dtype, (nmom,)),
                 ('EMLINE_SSB_1RE', self.float_dtype, (nmom,)),
                 ('EMLINE_SSB_PEAK', self.float_dtype, (nmom,)),
                 ('EMLINE_SEW_1RE', self.float_dtype, (nmom,)),
                 ('EMLINE_SEW_PEAK', self.float_dtype, (nmom,)),
#                 ('EMLINE_NAME', '<U{0:d}'.format(self.str_len['EMLINE_NAME']*neml)),
                 ('EMLINE_GFLUX_CEN', self.float_dtype, (neml,)),
                 ('EMLINE_GFLUX_1RE', self.float_dtype, (neml,)),
                 ('EMLINE_GFLUX_TOT', self.float_dtype, (neml,)),
                 ('EMLINE_GSB_1RE', self.float_dtype, (neml,)),
                 ('EMLINE_GSB_PEAK', self.float_dtype, (neml,)),
                 ('EMLINE_GEW_1RE', self.float_dtype, (neml,)),
                 ('EMLINE_GEW_PEAK', self.float_dtype, (neml,)),
#                 ('SPECINDEX_NAME', '<U{0:d}'.format(self.str_len['SPECINDEX_NAME']*nindx)),
#                 ('SPECINDEX_UNIT', '<U{0:d}'.format(self.str_len['SPECINDEX_UNIT']*nindx)),
                 ('SPECINDEX_LO', self.float_dtype, (nindx,)),
                 ('SPECINDEX_HI', self.float_dtype, (nindx,)),
                 ('SPECINDEX_LO_CLIP', self.float_dtype, (nindx,)),
                 ('SPECINDEX_HI_CLIP', self.float_dtype, (nindx,)),
                 ('SPECINDEX_1RE', self.float_dtype, (nindx,)),
                 ('SFR_1RE', self.float_dtype),
                 ('SFR_TOT', self.float_dtype)
               ]

    def _get_completed_observations(self):
        """
        Get the list of DRP completed (and prospectively DAP analyzed)
        observations.

        .. todo:
            Save drpc as an attribute of the class?

        Returns:
            numpy.ndarray: Two arrays listing the plate and ifudesign
            numbers of the available observations that the DAP should
            have analyzed.

        Raises:
            FileNotFoundError: Raised if the DRPComplete file is not
                available.
        """
        # Initialize the DRP complete file (read-only)
        drpc = DRPComplete(readonly=True, drpver=self.drpver, redux_path=self.redux_path,
                           dapver=self.dapver, analysis_path=self.analysis_path)
        if not os.path.isfile(drpc.file_path()):
            raise FileNotFoundError('Could not access DRP complete file: {0}'.format(
                                                                                drpc.file_path()))
    
        # Find the list of files that should have been analyzed
        indx = drpc.can_analyze()
        return drpc['PLATE'][indx], drpc['IFUDESIGN'][indx]

#    def _combine_plateifu_methods(self, plt, ifu):
#        """
#        Run the combinatorics to create the full list of plates,
#        ifudesigns, and methods.
#        """
#        nmethods = len(self.methods)
#        ncomplete = len(plt)
#        methods = numpy.array([ numpy.array(self.methods) ]*ncomplete).T.ravel()
#        plates = numpy.array([ numpy.array(plt) ]*nmethods).ravel() #.astype(str)
#        ifudesigns = numpy.array([ numpy.array(ifu) ]*nmethods).ravel() #.astype(str)
#        return methods, plates, ifudesigns

    def _construct_maps_file_list(self, method):
        """
        Construct the list of MAPS files that should be in/added to the
        DAPall file for a given method.
        """
        # Construct the file names
        return numpy.array(['{0}/{1}/{2}/{3}/manga-{2}-{3}-MAPS-{1}.fits.gz'.format(
                                self.analysis_path, method, p, i) 
                                    for p,i in zip(self.plate, self.ifudesign)])
    
    def _add_channels_to_header(self, hdr, channels, prefix, comment, units=None):
        """
        units is supposed to be a list or numpy array
        """
        ch = numpy.array(list(channels.values())).astype(int)
        srt = numpy.argsort(ch)
        ch = ch[srt]
        keys = numpy.array(list(channels.keys()))[srt]
        ndig = int(numpy.log10(ch[-1]))+1
        for c,k in zip(ch,keys):
            hdr[prefix+'{0}'.format(c+1).zfill(ndig)] = (k, comment)
            if units is not None:
                hdr[prefix+'U'+'{0}'.format(c+1).zfill(ndig)] = (units[c], comment+' unit')
        return hdr

    def _srd_snr_metric(self, dapmaps):
        """Grab the SNR metrics from the MAPS headers."""
        filters = [ 'G', 'R', 'I', 'Z' ]
        return numpy.array([dapmaps['PRIMARY'].header['SNR{0}MED'.format(f)] for f in filters]), \
                numpy.array([dapmaps['PRIMARY'].header['SNR{0}RING'.format(f)] for f in filters])

    @staticmethod
    def _radial_coverage_metric(r, ell):
        step = 1.0
        width = 2.5
        pixelscale = 0.5
        coverage_limit = 0.9

        maxr = numpy.ma.amax(r)
        binr = numpy.arange(width/2, maxr+width/4, width/4)
        coverage = numpy.zeros(len(binr), dtype=float)
        coverage[0] = numpy.sum(r < binr[0])/(numpy.pi*(1-ell)*numpy.square(binr[0]))
        for i in range(1,len(binr)):
            binrl = binr[i] - width/2
            binru = binrl + width
            ellipse_area = numpy.pi*(1-ell)*(numpy.square(binru) - numpy.square(binrl))
            coverage[i] = numpy.sum((r > binrl) & (r < binru))*numpy.square(pixelscale) \
                                / ellipse_area

        indx = numpy.arange(len(binr))[coverage > coverage_limit]
        if len(indx) == 0:
            warnings.warn('No ring had larger than 90% coverage!')
            return 0.
        interp = interpolate.interp1d(coverage[indx[-1]:], binr[indx[-1]:])
        return interp(coverage_limit)

    @staticmethod
    def _binning_metrics(dapmaps):
        """
        Return binning metrics:
            - Maximum radius of any valid bin
            - Number and median S/N of bins between 0-1, 0.5-1.5, and
              1.5-2.5 Re
        """
        # Unique bins
        # TODO: Select the BINID channel using the channel dictionary
        unique, indx = numpy.unique(dapmaps['BINID'].data[0,:,:].ravel(), return_index=True)
        if unique[0] == -1:
            unique = unique[1:]
            unique_indx = indx[1:]

        # Get the radii (in units of Re) of the unmasked, unique
        # bins constructed during the spatial binning
        r_re = dapmaps['BIN_LWELLCOO'].data.copy()[1,:,:].ravel()[unique_indx]

        # Get the S/N metric resulting from the spatial binning
        snr = dapmaps['BIN_SNR'].data.copy().ravel()[unique_indx]

        # Get the number of bins and median S/N between 0-1,
        # 0.5-1.5, and 1.5-2.5 Re
        rlim = numpy.array([ [0,1], [0.5,1.5], [1.5,2.5]])
        bin_r_n = numpy.empty(len(rlim), dtype=float)
        bin_r_snr = numpy.empty(len(rlim), dtype=float)
        for i in range(len(rlim)):
            indx = (r_re > rlim[i,0]) & (r_re < rlim[i,1])
            if numpy.sum(indx) == 0:
                bin_r_n[i] = 0
                bin_r_snr[i] = 0
                continue
            bin_r_n[i] = numpy.sum(indx)
            bin_r_snr[i] = numpy.median(snr[indx])
        
        return numpy.amax(r_re), bin_r_n, bin_r_snr

    def _stellar_kinematics_metrics(self, dapmaps, redshift):

        # Unique bins
        # TODO: Select the BINID channel using the channel dictionary
        unique, indx = numpy.unique(dapmaps['BINID'].data[1,:,:].ravel(), return_index=True)
        if unique[0] == -1:
            unique = unique[1:]
            unique_indx = indx[1:]

        # Pull the data from the maps file
        r_re = dapmaps['BIN_LWELLCOO'].data.copy()[1,:,:].ravel()[unique_indx]

        bin_flux = numpy.ma.MaskedArray(dapmaps['BIN_MFLUX'].data.copy(),
                                        mask=self.maps_bm.flagged(dapmaps['BIN_MFLUX_MASK'].data,
                                                                 'DONOTUSE')).ravel()[unique_indx]

        svel = numpy.ma.MaskedArray(dapmaps['STELLAR_VEL'].data.copy(),
                                    mask=self.maps_bm.flagged(dapmaps['STELLAR_VEL_MASK'].data,
                                                             'DONOTUSE')).ravel()[unique_indx]

        ssig = numpy.ma.MaskedArray(dapmaps['STELLAR_SIGMA'].data.copy(),
                                    mask=self.maps_bm.flagged(dapmaps['STELLAR_SIGMA_MASK'].data,
                                                              'DONOTUSE')).ravel()[unique_indx]

        ssig2corr = numpy.square(ssig) - numpy.square(
#                                    dapmaps['STELLAR_SIGMACORR'].data.ravel()[unique_indx])
                                    dapmaps['STELLAR_SIGMACORR'].data[0,:,:].ravel()[unique_indx])

        scchi = dapmaps['STELLAR_FOM'].data[2,:,:].copy().ravel()[unique_indx]

        # Convert velocities to redshift
        z = (svel*(1+redshift) + astropy.constants.c.to('km/s').value*redshift) \
                                    / astropy.constants.c.to('km/s').value

        # Get the bins within an on-sky circular aperture of 2.5 arcsec
        d2 = numpy.sum(numpy.square(dapmaps['BIN_LWSKYCOO'].data), axis=0).ravel()[unique_indx]
        center = d2 < 1.25*1.25

        # Get the flux-weighted mean velocity at the center
        wsum = numpy.ma.sum(bin_flux[center])
        stellar_z = -999. if numpy.all(bin_flux.mask[center] | z.mask[center]) \
                                or numpy.isclose(wsum, 0.) \
                            else numpy.ma.sum(bin_flux[center]*z[center])/wsum
        if not numpy.isfinite(stellar_z):
            raise ValueError('Stellar_z is not finite: {0}'.format(stellar_z))

        # Get the growth ranges of the stellar velocity
        stellar_vel_lo, stellar_vel_hi = sample_growth(svel.compressed(), [0.025,0.975])
        stellar_vel_lo_clip, stellar_vel_hi_clip = sample_growth(sigma_clip(svel).compressed(),
                                                                 [0.025,0.975])

        # Get the flux-weighted, corrected sigma within 1 Re
        within_1re = r_re < 1.
        wsum = numpy.ma.sum(bin_flux[within_1re])
        stellar_sigma2_1re = -999. if numpy.all(bin_flux.mask[within_1re]) \
                                            or numpy.isclose(wsum, 0.) \
                                else numpy.ma.sum(bin_flux[within_1re]*ssig2corr[within_1re])/wsum
        stellar_sigma_1re = numpy.sqrt(stellar_sigma2_1re) if stellar_sigma2_1re > 0 else -999.

        # Get the median reduced chi-square
        stellar_rchi2_1re = -999. if numpy.sum(within_1re) == 0 \
                                        else numpy.median(scchi[within_1re])

        return stellar_z, stellar_vel_lo, stellar_vel_hi, \
                        stellar_vel_lo_clip, stellar_vel_hi_clip, stellar_sigma_1re, \
                        stellar_rchi2_1re

    def _halpha_kinematics_metrics(self, dapmaps, deconstruct, redshift):

        # Unique bins
        # TODO: Select the BINID channel using the channel dictionary
        unique, indx = numpy.unique(dapmaps['BINID'].data[3,:,:].ravel(), return_index=True)
        if unique[0] == -1:
            unique = unique[1:]
            unique_indx = indx[1:]

        # Channel dictionary
        emline = channel_dictionary(dapmaps, 'EMLINE_GFLUX')

        # Pull the data from the maps file
        cooext = 'SPX_ELLCOO' if deconstruct else 'BIN_LWELLCOO'
        r_re = dapmaps[cooext].data.copy()[1,:,:].ravel()[unique_indx]

        try:
            flux = numpy.ma.MaskedArray(dapmaps['EMLINE_GFLUX'].data.copy()[emline['Ha-6564'],:,:],
                                        mask=self.maps_bm.flagged(
                                        dapmaps['EMLINE_GFLUX_MASK'].data[emline['Ha-6564'],:,:],
                                                              'DONOTUSE')).ravel()[unique_indx]
        except KeyError as e:
            print(e)
            warnings.warn('No halpha-data will be available.')
            return (-999.,)*8

        vel = numpy.ma.MaskedArray(dapmaps['EMLINE_GVEL'].data.copy()[emline['Ha-6564'],:,:],
                                   mask=self.maps_bm.flagged(
                                        dapmaps['EMLINE_GVEL_MASK'].data[emline['Ha-6564'],:,:],
                                                             'DONOTUSE')).ravel()[unique_indx]

        sig = numpy.ma.MaskedArray(dapmaps['EMLINE_GSIGMA'].data.copy()[emline['Ha-6564'],:,:],
                                   mask=self.maps_bm.flagged(
                                        dapmaps['EMLINE_GSIGMA_MASK'].data[emline['Ha-6564'],:,:],
                                                             'DONOTUSE')).ravel()[unique_indx]

        sig2corr = numpy.square(sig) - numpy.square(
                    dapmaps['EMLINE_INSTSIGMA'].data[emline['Ha-6564'],:,:].ravel()[unique_indx])

        chi = dapmaps['EMLINE_FOM'].data[2,:,:].copy().ravel()[unique_indx]

        # Convert velocities to redshift
        z = (vel*(1+redshift) + astropy.constants.c.to('km/s').value*redshift) \
                                    / astropy.constants.c.to('km/s').value

        # Get the bins within an on-sky circular aperture of 2.5 arcsec
        cooext = 'SPX_SKYCOO' if deconstruct else 'BIN_LWSKYCOO'
        d2 = numpy.sum(numpy.square(dapmaps[cooext].data), axis=0).ravel()[unique_indx]
        center = d2 < 1.25*1.25

        # Get the flux-weighted mean velocity at the center
        wsum = numpy.ma.sum(flux[center])

        halpha_z = -999. if numpy.all(flux.mask[center] | z.mask[center]) \
                                or numpy.isclose(wsum, 0.) \
                            else numpy.ma.sum(flux[center]*z[center])/wsum
        if not numpy.isfinite(halpha_z):
            raise ValueError('H-alpha z is not finite: {0}'.format(halpha_z))

#        halpha_z = -999. if numpy.all(flux.mask[center]) or numpy.isclose(wsum, 0.) \
#                            else numpy.ma.sum(flux[center]*z[center])/wsum

        # Get the growth ranges of the H-alpha velocity
        if numpy.all(vel.mask):
            halpha_gvel_lo = -999.
            halpha_gvel_hi = -999.
            halpha_gvel_lo_clip = -999.
            halpha_gvel_hi_clip = -999.
        else:
            halpha_gvel_lo, halpha_gvel_hi = sample_growth(vel.compressed(), [0.025,0.975])
            halpha_gvel_lo_clip, halpha_gvel_hi_clip = sample_growth(sigma_clip(vel).compressed(),
                                                                     [0.025,0.975])

        # Get the flux-weighted, corrected sigma within 1 Re
        within_1re = r_re < 1.
        wsum = numpy.ma.sum(flux[within_1re])
        halpha_gsigma2_1re = -999. if numpy.all(flux.mask[within_1re] | sig2corr.mask[within_1re]) \
                                        or numpy.isclose(wsum, 0.) \
                                    else numpy.ma.sum(flux[within_1re]*sig2corr[within_1re])/wsum
        if not numpy.isfinite(halpha_gsigma2_1re):
            raise ValueError('H-alpha gsigma^2 is not finite: {0}'.format(halpha_gsigma2_1re))
        halpha_gsigma_1re = numpy.sqrt(halpha_gsigma2_1re) if halpha_gsigma2_1re > 0 else -999.

        # Get the high-growth of the H-alpha velocity dispersion
        if numpy.all(sig2corr.mask):
            halpha_gsigma2_hi = -999.
            halpha_gsigma2_hi_clip = -999.
        else:
            halpha_gsigma2_hi = sample_growth(sig2corr.compressed(), 0.975)
            halpha_gsigma2_hi_clip = sample_growth(sigma_clip(sig2corr).compressed(), 0.975)

        halpha_gsigma_hi = numpy.sqrt(halpha_gsigma2_hi) if halpha_gsigma2_hi > 0 else -999.
        halpha_gsigma_hi_clip = numpy.sqrt(halpha_gsigma2_hi_clip) \
                                        if halpha_gsigma2_hi_clip > 0 else -999.

        # Get the median reduced chi-square for the emission-line module
        emline_rchi2_1re = -999. if numpy.sum(within_1re) == 0 \
                                        else numpy.median(chi[within_1re])
        

        return halpha_z, halpha_gvel_lo, halpha_gvel_hi, halpha_gvel_lo_clip, halpha_gvel_hi_clip, \
                        halpha_gsigma_1re, halpha_gsigma_hi, halpha_gsigma_hi_clip, \
                        emline_rchi2_1re

    def _emission_line_metrics(self, dapmaps, deconstruct, moment0=False):

        # Unique bins
        # TODO: Select the BINID channel using the channel dictionary
        unique, indx = numpy.unique(dapmaps['BINID'].data[3,:,:].ravel(), return_index=True)
        if unique[0] == -1:
            unique = unique[1:]
            unique_indx = indx[1:]

        flux_ext = 'EMLINE_SFLUX' if moment0 else 'EMLINE_GFLUX'
        neml = dapmaps[flux_ext].shape[0]

        # Pull the data from the maps file
        cooext = 'SPX_ELLCOO' if deconstruct else 'BIN_LWELLCOO'
        r_re = dapmaps[cooext].data.copy()[1,:,:].ravel()[unique_indx]

        flux = numpy.ma.MaskedArray(dapmaps[flux_ext].data.copy(),
                                    mask=self.maps_bm.flagged(dapmaps[flux_ext+'_MASK'].data,
                                                    'DONOTUSE')).reshape(neml,-1)[:,unique_indx]

        ew_ext = 'EMLINE_SEW' if moment0 else 'EMLINE_GEW'
        ew = numpy.ma.MaskedArray(dapmaps[ew_ext].data.copy(),
                                  mask=self.maps_bm.flagged(dapmaps[ew_ext+'_MASK'].data,
                                                    'DONOTUSE')).reshape(neml,-1)[:,unique_indx]

        # Get the total flux at the center
        cooext = 'SPX_SKYCOO' if deconstruct else 'BIN_LWSKYCOO'
        d2 = numpy.sum(numpy.square(dapmaps[cooext].data), axis=0).ravel()[unique_indx]
        center = d2 < 1.25*1.25
        if numpy.sum(center) == 0:
            warnings.warn('No data near center!')
            emline_flux_cen = numpy.full(neml, -999., dtype=float)
        else:
            emline_flux_cen = numpy.ma.sum(flux[:,center], axis=1).filled(-999.)
            
        # Get the total flux, mean surface-brightness, and mean
        # equivalent width within 1 Re
        within_1re = r_re < 1.
        if numpy.sum(within_1re) == 0:
            warnings.warn('No data within 1 Re!')
            emline_flux_1re = numpy.full(neml, -999., dtype=float)
            emline_sb_1re = numpy.full(neml, -999., dtype=float)
            emline_ew_1re = numpy.full(neml, -999., dtype=float)
        else:
            emline_flux_1re = numpy.ma.sum(flux[:,within_1re], axis=1).filled(-999.)
            emline_sb_1re = numpy.ma.mean(flux[:,within_1re], axis=1).filled(-999.)
            emline_ew_1re = numpy.ma.mean(ew[:,within_1re], axis=1).filled(-999.)

        # Get the total flux within the full field-of-fiew
        emline_flux_tot = numpy.ma.sum(flux, axis=1).filled(-999.)

        # Get the peak surface-brightness and equivalent width within
        # the full field-of-fiew
        emline_sb_peak = numpy.ma.amax(flux, axis=1).filled(-999.)
        emline_ew_peak = numpy.ma.amax(ew, axis=1).filled(-999.)

        return emline_flux_cen, emline_flux_1re, emline_flux_tot, emline_sb_1re, \
                        emline_sb_peak, emline_ew_1re, emline_ew_peak

    def _spectral_index_metrics(self, dapmaps, deconstruct, specindex_units):

        # Unique bins
        # TODO: Select the BINID channel using the channel dictionary
        unique, indx = numpy.unique(dapmaps['BINID'].data[4,:,:].ravel(), return_index=True)
        if unique[0] == -1:
            unique = unique[1:]
            unique_indx = indx[1:]

        nindx = dapmaps['SPECINDEX'].shape[0]

        # Pull the data from the maps file
        cooext = 'SPX_ELLCOO' if deconstruct else 'BIN_LWELLCOO'
        r_re = dapmaps[cooext].data.copy()[1,:,:].ravel()[unique_indx]
        specindex = numpy.ma.MaskedArray(dapmaps['SPECINDEX'].data.copy(),
                                         mask=self.maps_bm.flagged(dapmaps['SPECINDEX_MASK'].data,
                                                    'DONOTUSE')).reshape(nindx,-1)[:,unique_indx]
        specindex_corr = numpy.ma.MaskedArray(
                            dapmaps['SPECINDEX_CORR'].data.copy().reshape(nindx,-1)[:,unique_indx],
                                              mask=specindex.mask.copy())

        # Get the corrected indices; Unitless and ang units are treated
        # the same.
        mag = specindex_units == 'mag'
        ang = numpy.invert(mag)
        specindex_corr[ang] = specindex[ang]*specindex_corr[ang]
        specindex_corr[mag] = specindex[mag]+specindex_corr[mag]

        # Get the growth limits
        specindex_lo = numpy.empty(nindx, dtype=float)
        specindex_hi = numpy.empty(nindx, dtype=float)
        specindex_lo_clip = numpy.empty(nindx, dtype=float)
        specindex_hi_clip = numpy.empty(nindx, dtype=float)
        for i in range(nindx): 
            if numpy.all(specindex_corr.mask[i,:]):
                specindex_lo[i] = -999.
                specindex_hi[i] = -999.
                specindex_lo_clip[i] = -999.
                specindex_hi_clip[i] = -999.
                continue
            specindex_lo[i], specindex_hi[i] = sample_growth(specindex_corr[i,:].compressed(),
                                                             [0.025,0.975])
            specindex_lo_clip[i], specindex_hi_clip[i] \
                        = sample_growth(sigma_clip(specindex_corr[i,:]).compressed(), [0.025,0.975])

        # Get the median within 1 Re
        specindex_1re = numpy.ma.median(specindex[:,r_re < 1.], axis=1).filled(-999.)

        return specindex_lo, specindex_hi, specindex_lo_clip, specindex_hi_clip, specindex_1re

    def file_name(self):
        """Return the name of the DRP complete database."""
        name = defaults.dapall_file(drpver=self.drpver, dapver=self.dapver,
                                    analysis_path=self.analysis_path)
        return os.path.split(name)[1]

    def file_path(self):
        """Return the full pat to the DRP complete database."""
        return defaults.dapall_file(drpver=self.drpver, dapver=self.dapver,
                                    analysis_path=self.analysis_path)

    def method_database(self, method, deconstruct, drpall):
        """
        Construct the DAPall database for the provided DAP method.

        Args:
            method (:obj:`str`):
                Name of the DAP method
            deconstruct (:obj:`bool`):
                Flag that the method deconstructs the bins during the
                emission-line fitting module.
            drpall (`numpy.recarray`_):
                Primary DRPall data table

        Returns:
            `numpy.recarray`_: DAPall database.
        """
        # Construct the list of files to look for
        dapfiles = self._construct_maps_file_list(method)

        # Initialize the output database
        db = init_record_array(self.ndap, self._table_dtype(self.nmom, self.neml, self.nindx))

        # Add the basic information
        db['PLATE'] = self.plate
        db['IFUDESIGN'] = self.ifudesign
        db['PLATEIFU'] = numpy.array(['{0}-{1}'.format(p,i) 
                                            for p,i in zip(self.plate, self.ifudesign)])
        # For now only analyzing cubes!
        db['MODE'][:] = 'CUBE'
        db['DAPTYPE'][:] = method

        # Copy from DRPall
        columns_from_drpall = ['MANGAID', 'OBJRA', 'OBJDEC', 'IFURA', 'IFUDEC', 'MNGTARG1',
                               'MNGTARG2', 'MNGTARG3', 'NSA_Z', 'NSA_ZDIST',
                               'NSA_ELPETRO_BA', 'NSA_ELPETRO_PHI', 'NSA_ELPETRO_TH50_R',
                               'NSA_SERSIC_BA', 'NSA_SERSIC_PHI', 'NSA_SERSIC_TH50',
                               'NSA_SERSIC_N', 'VERSDRP2', 'VERSDRP3', 'VERSCORE', 'VERSUTIL',
                               'DRP3QUAL']

        columns_from_maps_hdr = ['VERSDAP', 'DAPQUAL', 'RDXQAKEY', 'BINKEY', 'SCKEY', 'ELMKEY',
                                 'ELFKEY', 'SIKEY', 'BINTYPE']

        # Itereate through each maps file
        for i in range(self.ndap):
            print('Processing {0}; observation {1}/{2}'.format(method, i+1,self.ndap))#, end='\r')

            # Find the index in the drpall file
            indx = numpy.where(drpall['PLATEIFU'] == db['PLATEIFU'][i])[0]
            if len(indx) > 1:
                warnings.warn('More than one entry in DRPall file: {0}.'.format(db['PLATEIFU'][i]))
            if len(indx) != 1:
                indx = [-1]
                warnings.warn('Could not find {0} in DRPall file!'.format(db['PLATEIFU'][i]))
            db['DRPALLINDX'][i] = indx[0]

            # Add data from the DRPall file
            if db['DRPALLINDX'][i] != -1:
                for c in columns_from_drpall:
                    db[c][i] = drpall[c][db['DRPALLINDX'][i]]

            # Set the cosmological distances based on the NSA redshift
            db['LDIST_NSA_Z'][i] = self.cosmo.luminosity_distance(db['NSA_Z'][i]).value \
                                        if db['NSA_Z'][i] > 0 else -999.
            db['ADIST_NSA_Z'][i] = self.cosmo.angular_diameter_distance(db['NSA_Z'][i]).value \
                                        if db['NSA_Z'][i] > 0 else -999.

            # Open the maps file
            if not os.path.isfile(dapfiles[i]):
                db['DAPDONE'][i] = 0    # A fault occurred such that the MAPS file was not produced
                continue
            db['DAPDONE'][i] = 1        # DAP successfully produced the file
            dapmaps = fits.open(dapfiles[i])

            # Add information from the DAP MAPS file header
            for c in columns_from_maps_hdr:
                db[c][i] = dapmaps['PRIMARY'].header[c]

            # Additional elements from the header that need to be handled
            # or have different names
            # TODO: REMOVE all of the latter to make sure keywords match
            try:
                db['BINSNR'][i] = dapmaps['PRIMARY'].header['BINSNR']
            except:
                db['BINSNR'][i] = -999.0
            db['TPLKEY'][i] = dapmaps['PRIMARY'].header['PPXFTPLK']
            db['DAPBINS'][i] = dapmaps['PRIMARY'].header['NBINS']
            try:
                db['DATEDAP'][i] = dapmaps['PRIMARY'].header['DATEDAP']
            except:
                db['DATEDAP'][i] = time.strftime('%Y-%m-%d',
                                                 time.localtime(os.path.getmtime(dapfiles[i])))

            # TODO: Add any checks that the DRPall data matches the
            # header data in the MAPS files?

            # Basic mask for all maps: mask any spaxel not included
            # in any bin
            basic_map_mask = dapmaps['BINID'].data[0,:,:] < 0

            # Set input redshift and cosmological distances based on
            # this redshift
            db['Z'][i] = dapmaps['PRIMARY'].header['SCINPVEL']/astropy.constants.c.to('km/s').value
            db['LDIST_Z'][i] = self.cosmo.luminosity_distance(db['Z'][i]).value
            db['ADIST_Z'][i] = self.cosmo.angular_diameter_distance(db['Z'][i]).value

            # Determine the coverage of the datacube (independent of the
            # DAP method)
            r = numpy.ma.MaskedArray(dapmaps['SPX_ELLCOO'].data.copy()[0,:,:], mask=basic_map_mask)
            db['RCOV90'][i] = self._radial_coverage_metric(r.compressed(),
                                                           1-db['NSA_ELPETRO_BA'][i])

            # Get the SRD-based S/N metric (independent of the DAP
            # method)
            r_re = numpy.ma.MaskedArray(dapmaps['SPX_ELLCOO'].data.copy()[1,:,:],
                                        mask=numpy.invert(dapmaps['SPX_SNR'].data > 0))
            db['SNR_MED'][i], db['SNR_RING'][i] = self._srd_snr_metric(dapmaps)

            # Get the mean surface brightness within 1 Re used by the
            # stellar-continuum fitting (independent of the DAP method)
            mflux = numpy.ma.MaskedArray(dapmaps['SPX_MFLUX'].data.copy(),
                                         mask=numpy.invert(dapmaps['SPX_SNR'].data > 0))
            within_1re = r_re < 1
            if numpy.sum(within_1re) == 0:
                warnings.warn('No data within 1 Re!')
                db['SB_1RE'][i] = -999.
            else:
                db['SB_1RE'][i] = numpy.ma.mean(mflux[within_1re])

            # Get the spatial binning metrics
            db['BIN_RMAX'][i], db['BIN_R_N'][i], db['BIN_R_SNR'][i] \
                            = self._binning_metrics(dapmaps)

            # Get the metrics for the stellar kinematics
            db['STELLAR_Z'][i], db['STELLAR_VEL_LO'][i], db['STELLAR_VEL_HI'][i], \
                    db['STELLAR_VEL_LO_CLIP'][i], db['STELLAR_VEL_HI_CLIP'][i], \
                    db['STELLAR_SIGMA_1RE'][i], db['STELLAR_RCHI2_1RE'][i] \
                            = self._stellar_kinematics_metrics(dapmaps, db['Z'][i])

            # Get the metrics for the emission-line kinematics
            db['HA_Z'][i], db['HA_GVEL_LO'][i], db['HA_GVEL_HI'][i], db['HA_GVEL_LO_CLIP'][i], \
                    db['HA_GVEL_HI_CLIP'][i], db['HA_GSIGMA_1RE'][i], db['HA_GSIGMA_HI'][i], \
                    db['HA_GSIGMA_HI_CLIP'][i], db['EMLINE_RCHI2_1RE'][i] \
                            = self._halpha_kinematics_metrics(dapmaps, deconstruct, db['Z'][i])

            # Get the moment-based emission-line metrics
            db['EMLINE_SFLUX_CEN'][i], db['EMLINE_SFLUX_1RE'][i], db['EMLINE_SFLUX_TOT'][i], \
                    db['EMLINE_SSB_1RE'][i], db['EMLINE_SSB_PEAK'][i], db['EMLINE_SEW_1RE'][i], \
                    db['EMLINE_SEW_PEAK'][i] \
                            = self._emission_line_metrics(dapmaps, deconstruct, moment0=True)

            # Get the Gaussian-based emission-line metrics
            db['EMLINE_GFLUX_CEN'][i], db['EMLINE_GFLUX_1RE'][i], db['EMLINE_GFLUX_TOT'][i], \
                    db['EMLINE_GSB_1RE'][i], db['EMLINE_GSB_PEAK'][i], db['EMLINE_GEW_1RE'][i], \
                    db['EMLINE_GEW_PEAK'][i] = self._emission_line_metrics(dapmaps, deconstruct)

            # Get the spectral-index metrics
            db['SPECINDEX_LO'][i], db['SPECINDEX_HI'][i], db['SPECINDEX_LO_CLIP'][i], \
                    db['SPECINDEX_HI_CLIP'][i], db['SPECINDEX_1RE'][i] \
                            = self._spectral_index_metrics(dapmaps, deconstruct, self.spindx_units)

            # Estimate the star-formation rate
            log_Mpc_in_cm = numpy.log10(astropy.constants.pc.to('cm').value) + 6
            log_halpha_luminosity_1re = numpy.log10(4*numpy.pi) \
                        + numpy.ma.log10(db['EMLINE_GFLUX_1RE'][i,self.elfit_channels['Ha-6564']]) \
                        - 17 + 2*numpy.ma.log10(db['LDIST_Z'][i]) + 2*log_Mpc_in_cm
            db['SFR_1RE'][i] = numpy.ma.power(10, log_halpha_luminosity_1re - 41.27).filled(-999.)
            log_halpha_luminosity_tot = numpy.log10(4*numpy.pi) \
                        + numpy.ma.log10(db['EMLINE_GFLUX_TOT'][i,self.elfit_channels['Ha-6564']]) \
                        - 17 + 2*numpy.ma.log10(db['LDIST_Z'][i]) + 2*log_Mpc_in_cm
            db['SFR_TOT'][i] = numpy.ma.power(10, log_halpha_luminosity_tot - 41.27).filled(-999.)

        print('Processing {0}; observation {1}/{1}'.format(method, self.ndap))

        # Check that all the data is finite.
        found = False
        for name in db.dtype.names:
            if not (numpy.issubdtype(db[name].dtype, numpy.integer) \
                        or numpy.issubdtype(db[name].dtype, numpy.floating)):
                continue
            if not numpy.all(numpy.isfinite(db[name])):
                print('All values not finite in column {0} for method {1}'.format(name, method))
                found = True
        if found:
            raise ValueError('All values not finite in DAPall table.')

        return db

    def update(self, plan, methods=None):
        """
        Update the DAPall file.

        Currently, this constructs the entire DAPall file from
        scratch, regardless of whether or not some fraction of the
        output files have already been analyzed.
        
        Args:
            plan (:class:`mangadap.par.analysisplan.AnalysisPlanSet`):
                The plan object used by the DAP.
            methods (:obj:`str`, :obj:`list`, optional):
                Specify a set of methods in the DAP plan file to
                include in the DAPall file. If not provided, all
                methods in the plan file are included.

        Raises:
            ValueError:
                Raised if a provided method is not in the provided
                set of analysis plans, or if the methods do not all
                have the same number of emission-line bandpasses,
                emission-lines to fit, and number of spectral
                indices.
        """
        if self.readonly:
            raise ValueError('Object is read-only.')

        # Check the input type
        # TODO: Plan isn't really needed.  Could just troll the
        # analysis_path for the directory names...
        if not isinstance(plan, AnalysisPlanSet):
            raise TypeError('Input plan must have type AnalysisPlanSet.')

        # Get the full list of available plan methods
        plan_methods = []
        deconstruct = []
        for p in plan:
            bin_method = SpatiallyBinnedSpectra.define_method(p['bin_key'])
            sc_method = StellarContinuumModel.define_method(p['continuum_key'])
            el_method = EmissionLineModel.define_method(p['elfit_key'])
            plan_methods += [defaults.dap_method(bin_method['key'],
                                                 sc_method['fitpar']['template_library_key'],
                                                 el_method['continuum_tpl_key'])]
            deconstruct += [el_method['deconstruct_bins'] != 'ignore']

        # Check that the plan methods are unique.  This should never be
        # raised, unless it also causes problems in the DAP itself.
        if len(numpy.unique(plan_methods)) != len(plan_methods):
            raise ValueError('All of the plan methods must be unique!')

        # Get the list of methods to look for
        if methods is None:
            self.methods = numpy.array(plan_methods)
            self.deconstruct = numpy.array(deconstruct)
            use_plans = numpy.ones(len(plan_methods), dtype=bool)
        else:
            # Make sure the provided list is unique
            if len(numpy.unique(methods)) != len(methods):
                raise ValueError('Must provide unique list of methods.')
            # All the methods to use must be in the provided plan
            if not numpy.all( [m in plan_methods for m in methods ]):
                raise ValueError('All provided methods must be in provided plan file.')

            # Find the index of the plans to use
            # TODO: Re-write this as a list compression
            self.methods = []
            self.deconstruct = []
            use_plans = numpy.zeros(len(plan_methods), dtype=bool)
            for i in range(len(plan_methods)):
                if plan_method[i] in methods:
                    use_plans[i] = True
                    self.methods += [plan_method[i]]
                    self.deconstruct += [deconstruct[i]]
            self.methods = numpy.array(self.methods)
            self.deconstruct = numpy.array(self.deconstruct)

        # Check that all the plans to use have the same emission-line
        # bandpass channels, ...
        elmom_keys, elmom_channels \
                    = numpy.array([DAPall._emission_line_moment_db_info(p['elmom_key'])
                                        for p in plan[use_plans]], dtype=object).T
        if numpy.any([ len(k) != len(elmom_channels[0]) or len(set(k)-set(elmom_channels[0])) > 0
                        for k in elmom_channels]):
            raise ValueError('All plans must provide the same emission-line bandpass channels.')
        self.elmom_channels = elmom_channels[0].copy()
        del elmom_channels
        self.nmom = len(self.elmom_channels)

        # ..., emission-line fit channels, and...
        elfit_keys, elfit_channels \
                    = numpy.array([DAPall._emission_line_db_info(p['elfit_key'])
                                        for p in plan[use_plans]], dtype=object).T
        if numpy.any([ len(k) != len(elfit_channels[0]) or len(set(k)-set(elfit_channels[0])) > 0
                        for k in elfit_channels]):
            raise ValueError('All plans must provide the same emission-line fit channels.')
        self.elfit_channels = elfit_channels[0].copy()
        del elfit_channels
        self.neml = len(self.elfit_channels)

        # ..., and spectral-index channels
        abs_keys, bhd_keys, spindx_channels, spindx_units \
                    = numpy.array([DAPall._spectral_index_db_info(p['spindex_key'])
                                        for p in plan[use_plans]], dtype=object).T
        if numpy.any([ len(k) != len(spindx_channels[0]) or len(set(k)-set(spindx_channels[0])) > 0
                        for k in spindx_channels]):
            raise ValueError('All plans must provide the same emission-line fit channels.')
        self.spindx_channels = spindx_channels[0].copy()
        self.spindx_units = spindx_units[0].copy()
        del spindx_channels, spindx_units
        self.nindx = len(self.spindx_channels)

        # PlateIFUs that should be analyzed per DAP method
        # TODO: Somehow sort these?
        self.plate, self.ifudesign = self._get_completed_observations()
        # Number of completed observations per analysis method
        self.ndap = len(self.plate)

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, '{0:^50}'.format('RUNNING DAPALL UPDATE'))
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO,
                       'Completed observations: {0}'.format(self.ndap))
            log_output(self.loggers, 1, logging.INFO,
                       'Available methods: {0}'.format(plan_methods))
            log_output(self.loggers, 1, logging.INFO,
                       'Methods for output: {0}'.format(self.methods))
            log_output(self.loggers, 1, logging.INFO,
                       'Emission line moments: {0}'.format(self.nmom))
            log_output(self.loggers, 1, logging.INFO,
                       'Emission lines fit: {0}'.format(self.neml))
            log_output(self.loggers, 1, logging.INFO,
                       'Spectral indices: {0}'.format(self.nindx))

        # If the HDUList is already initialized, close and reinitialize
        # it
        if self.hdu is not None:
            self.hdu.close()
            self.hdu = None

        # Open the DRPall file
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                       'Reading DRPall: {0}'.format(self.drpall_file))
        hdu_drpall = fits.open(self.drpall_file)
        drpall = hdu_drpall[1].data

        # Create the primary header
        hdr = fits.Header()
        hdr['DATE'] = (time.strftime('%Y-%m-%d',time.gmtime()), 'UTC date created')
        hdr['VERSDRP3'] = (self.drpver, 'DRP version')
        hdr['VERSDAP'] = (self.dapver, 'DAP version')

        # Add the emission-line moment channel header
        hdr = self._add_channels_to_header(hdr, self.elmom_channels, 'ELS',
                                           'Summed emission-line element')
        hdr = self._add_channels_to_header(hdr, self.elfit_channels, 'ELG',
                                           'Gaussian fit emission-line element')
        hdr = self._add_channels_to_header(hdr, self.spindx_channels, 'SPI',
                                           'Spectral-indx element', units=self.spindx_units)

        # Construct a DAPall database for each analysis method and add
        # it as a binary table to the list of HDUs.
        self.hdu = [fits.PrimaryHDU(header=hdr)]
        for method, deconstruct in zip(self.methods, self.deconstruct):
            method_db = self.method_database(method, deconstruct, drpall)
            self.hdu += [fits.BinTableHDU.from_columns(
                                [fits.Column(name=n, format=rec_to_fits_type(method_db[n]),
                                             array=method_db[n]) for n in method_db.dtype.names],
                                                       name=method)]
        self.hdu = fits.HDUList(self.hdu)

        # Write the file (use of this function is probably overkill)
        DAPFitsUtil.write(self.hdu, self.file_path(), overwrite=True, checksum=True,
                          loggers=self.loggers)
        




"""
Implements a base class that performs linear least-squares fitting of
1D basis functions constructed via a pseudo-Vandermonde matrix.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import os
import time

from IPython import embed

import numpy

from . import modeling

# TODO: Generalize this even further so that a user can provide their
# own "basis" functions. And change the name of this module to
# basis.py?

class Vander1D:
    r"""
    Base class for least-squares fitting using linear combinations of 1D
    basis functions.

    Args:
        vander (callable):
            Function used to generate the pseudo-Vandermonde matrix
            of the basis functions; e.g.,
            `numpy.polynomial.legendre.legvander`_. Calling sequence
            must be ``vander(x,order)``.
        x (array-like):
            Ordinate for fitting.
        order (:obj:`int`):
            Order of the fit.
        rcond (:obj:`float`, optional):
            Relative condition number of the fit. Singular values
            smaller than this relative to the largest singular value
            will be ignored. The default value is len(x)*eps, where eps
            is the relative precision of the float type, about 2e-16 in
            most cases.

    Attributes:
        x (`numpy.ndarray`_):
            See the instantiation argument.  Shape is :math:`(N_{\rm
            x},)`.
        v (`numpy.ndarray`_):
            Pseudo-Vandermonde matrix of the request fitting order,
            :math:`o`.  Shape is :math:`(o+1, N_{\rm x})`.
        rcond (:obj:`float`):
            See the instantiation argument.
    """

    def __init__(self, vander, x, order, rcond=None):
        self.x = numpy.atleast_1d(x)
        if self.x.ndim != 1:
            raise ValueError('Ordinate must be a vector.')

        self.v = vander(self.x, order).T
        self.rcond = self.x.size * numpy.finfo(self.x.dtype).eps if rcond is None else rcond

    @property
    def size(self):
        """The size of the coordinate vector."""
        return self.x.size

    @property
    def shape(self):
        """The shape of the coordinate vector."""
        return self.x.shape

    @property
    def order(self):
        """The order of the function."""
        return self.v.shape[0]-1

    def fit(self, y, w=None, err=None, mask=None, rej_iter=0, rej_lo=3., rej_hi=3., rej_win=None,
            model=False):
        r"""
        Fit the provided data.

        The ordinate is provided on the class instantiation.

        If provided, the errors, :math:`\epsilon`, are converted to
        weights to use during the fitting, where :math:`\omega_i =
        1/\epsilon_i`. Both weights, :math:`w`, and errors can be
        provided; however, in this case, the weights used during the
        fit are are :math:`\omega_i = w_i / \epsilon_i`.

        .. todo::
            
            Allow for the nominal calculation of the errors in the
            coefficients.

        Args:
            y (array-like):
                A 1D or 2D array with the 1D functions to fit. Shape
                is :math:`(N_x,)` or :math:`(N_{\rm vec},N_x)`, where
                :math:`N_x` is the length of the ordinate
                (:attr:`x`). Input can be a masked array.
            w (array-like, optional):
                A 1D or 2D array with the weights for each pixel.
                Shape is :math:`(N_x,)` or :math:`(N_{\rm vec},N_x)`,
                where :math:`N_x` is the length of the ordinate
                (:attr:`x`). If ``w`` is 1D, but ``y`` is 2D, the
                same weights are applied to all vectors.
            err (array-like, optional):
                Error in ``y``, used for weighting. Shape must match
                ``y``.
            mask (array-like, optional):
                A boolean mask; values to ignore should have ``mask
                == True``. Shape must match ``y``. Treated
                identically as a mask provided by passing in a
                `numpy.ma.MaskedArray`_ for ``y``. Note that
                providing a weight of 0 is identical to masking the
                value.
            rej_iter (:obj:`int`, optional):
                Number of fit rejection iterations.  If 0, no rejection
                iterations are performed.  If -1, rejections are
                performed until no points are rejected in a given
                iteration.
            rej_lo (scalar-like, optional):
                The lower sigma rejection threshold.
            rej_hi (scalar-like, optional):
                The higher sigma rejection threshold.
            rej_win (:obj:`int`, optional):
                The length of a boxcar window to use in a calculation of
                a *local* standard deviation used during the rejection
                iterations.  This is useful if the S/N varies strongly
                within the provided vector being fit.  If None, no
                windowing is applied, meaning a *global* standard
                deviation is used for rejection.
            model (:obj:`bool`, optional):
                If True, return the model as well as the best fitting
                coefficients.

        Returns:
            `numpy.ndarray`, :obj:`tuple`: As many as three objects
            can be returned. Regardless of the input, the function
            always returns an array with the best-fitting
            coefficients, with shape :math:`(o+1,)` or :math:`(N_{\rm
            vec},o+1)`, depending on the input shape of ``y``. If
            ``model`` is True, an array with the best-fitting models
            are also returned (same shape as ``y``). If rejection
            operations are performed, a boolean array flagging the
            pixels excluded from the fit is also returned. Note that
            any data that will have been ignored based on the input
            (the element is masked or has 0 weight) is *not* flagged
            as having been rejected.
        """
        # Check the input vectors to fit
        _y = numpy.atleast_1d(y)
        if _y.ndim != 1 and _y.ndim != 2:
            raise ValueError('Input y coordinates must be a single vector or 2D array.')
        if _y.shape[-1] != self.x.size:
            raise ValueError('The last dimension of y must have the same length as the ordinate.')

        # Setup the weights
        if w is None:
            # Weights are needed if rejecting
            _w = None if rej_iter == 0 else numpy.ones(_y.shape, dtype=float) 
        else:
            _w = numpy.atleast_1d(w)
            if _w.ndim != 1 and _w.ndim != 2:
                raise ValueError('Input weights must be a single vector or 2D array.')
            if _w.ndim == 1 and _w.shape != self.x.shape:
                raise ValueError('If a single vector, the weights must have the same shape as '
                                 'the ordinate vector.')
            if rej_iter != 0 and _w.ndim != _y.ndim:
                # If rejecting, weights must be applied to each vector
                # separately
                _w = numpy.tile(_w, (_y.shape[0],1))
            if _w.ndim == 2 and _w.shape != _y.shape:
                raise ValueError('If an array, the weights must have the same shape as the data '
                                 'to fit.')
            if numpy.any(_w < 0):
                raise ValueError('Weights must be positive!')
        if isinstance(_y, numpy.ma.MaskedArray):
            _m = numpy.logical_not(numpy.ma.getmaskarray(_y)).astype(float)
            _w = _m if _w is None else _m * _w
            _y = _y.filled(0.0)
        if mask is not None:
            _m = numpy.logical_not(numpy.atleast_1d(mask)).astype(float)
            if _m.shape != _y.shape:
                raise ValueError('Mask must have the same shape as the data to fit.')
            _w = _m if _w is None else _m * _w

        # Check the input error and add it to the weights, if provided
        if err is not None:
            # Below is faster than numpy.ma.power(err, -1).filled(0.0)
            # and avoids errors when err is 0.
            _e = err.filled(0.0) if isinstance(err, numpy.ma.MaskedArray) else err
            _e = (_e > 0)/(_e + (_e == 0.))
            if _e.shape != _y.shape:
                raise ValueError('Input errors must have the same shape as the data to fit.')
            _w = _e if _w is None else _e * _w

        # Fit the data
        c = self._single_fit(_y, _w)

        # If no rejection iterations, we're done
        if rej_iter == 0:
            return (c, self.model(c)) if model else c

        # Initial mask
        init_good = _w > 0

        # Iteratively reject
        i = 0
        while i < rej_iter or rej_iter < 0:
            # Setup a masked array to hold the weighted residuals
            resid = _w * numpy.ma.MaskedArray(_y - self.model(c), mask=numpy.logical_not(_w > 0))
            # Reject pixels
            rej = modeling.reject_residuals_1d(resid, lo=rej_lo, hi=rej_hi, boxcar=rej_win)
            # Determine if any were rejected
            refit = numpy.any(rej, axis=-1)
            if not numpy.any(refit):
                # None were rejected so we're done
                break
            # Set the weight to 0 for the rejected data
            _w *= numpy.logical_not(rej).astype(float)
            # Refit
            c = self._single_fit(_y, _w)
            # Increment iteration
            i += 1

        rejected = numpy.logical_not(_w > 0) & init_good
        return (c, self.model(c), rejected) if model else (c, rejected)

    def model(self, c):
        r"""
        Return the model function constructed using the provided
        coefficients.

        Args:
            c (array-like):
                The coefficients for the model with shape shape
                :math:`(o+1,)` or :math:`(N_{\rm vec},o+1)`, where
                :math:`o` is the order of the function.

        Returns:
            `numpy.ndarray`_: Array with the evaluated functions. The
            output shape is :math:`(N_x,)` or :math:`(N_{\rm
            vec},N_x)` depending on the input shape of ``c``.
        """
        return numpy.dot(c, self.v)

    def _single_fit(self, y, w):
        r"""
        Perform a single fitting iteration.

        Args:
            y (`numpy.ndarray`_):
                The data to fit.  Shape is :math:`(N_{\rm x},)` or
                :math:`(N_{\rm vec}, N_{\rm x})`.  If the latter, a fit
                is done to each vector.
            w (`numpy.ndarray`_):
                Weights.  Shape is :math:`(N_{\rm x},)` or
                :math:`(N_{\rm vec}, N_{\rm x})`.  If the latter, this
                call is equivalent to looping over :math:`N_{\rm vec}`
                and fitting each vector in ``y`` individually.  Can be
                None, and if so the fit is unweighted.
        
        Returns:
            `numpy.ndarray`_: The best-fitting coefficients for the
            basis functions with order :math:`o`.  Shape is
            :math:`(o+1,)` or :math:`(N_{\rm vec},o+1)`.
        """
        if w is None:
            # Unweighted fit. For multiple y vectors, make sure
            # returned array has the shape described by the return
            # statement above.
            return numpy.linalg.lstsq(self.v.T, y.T, rcond=self.rcond)[0].T
        
        if y.ndim > 1 and w.shape == y.shape:
            # If weighting, can only solve all fits simultaneously if
            # the same weights are used for all vectors. Therefore,
            # need to loop through all y vectors here to accommodate
            # weights that may be different fro each y vector.
            c = numpy.zeros((y.shape[0], self.v.shape[0]), dtype=float)
            for i in range(y.shape[0]):
                c[i] = self._single_fit(y[i], w[i])
            return c

        # w is a single vector, used when fitting one or more y
        # vectors.
        lhs = self.v * w
        rhs = y * w
        scl = numpy.sqrt(numpy.square(lhs).sum(axis=1))
        scl[scl == 0] == 1
        c = numpy.linalg.lstsq(lhs.T/scl, rhs.T, rcond=self.rcond)[0].T
        return c/scl


class Legendre1D(Vander1D):
    """
    Least-squares fitting of 1D Legendre functions.

    Args:
        x (array-like):
            Ordinate for fitting.
        order (:obj:`int`):
            Order of the fit.
        rcond (:obj:`float`, optional):
            Relative condition number of the fit. Singular values
            smaller than this relative to the largest singular value
            will be ignored. The default value is len(x)*eps, where eps
            is the relative precision of the float type, about 2e-16 in
            most cases.
    """
    def __init__(self, x, order, rcond=None):
        super(Legendre1D, self).__init__(numpy.polynomial.legendre.legvander, x, order,
                                         rcond=rcond)




#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc, colors

from astropy.io import fits
import astropy.constants

from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.98')
    ax.grid(True, which='major', color='0.9', zorder=0, linestyle='-')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax

def mass_lha(drpall, dapall):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == 'HYB10-GAU-MILESHC')
    indx &= (drpall['nsa_sersic_mass'][dapall['DRPALLINDX']] > 0)

    indx &= (dapall['EMLINE_GSB_1RE'][:,11] > -999)
    indx &= (dapall['EMLINE_GSB_1RE'][:,18] > -999)
    indx &= (dapall['SFR_1RE'] > -999)

    decrement = numpy.ma.divide(dapall['EMLINE_GSB_1RE'][:,18],
                                   dapall['EMLINE_GSB_1RE'][:,11])

    loew_indx =  indx & (dapall['EMLINE_GEW_1RE'][:,18] > -999)\
                        & (dapall['EMLINE_GEW_1RE'][:,18] < 2)
    hiew_indx = indx & (dapall['EMLINE_GEW_1RE'][:,18] > 2)

    mass_lim = [1e8, 1e12]
    lha_lim = [36.5, 42.5]

    font = { 'size' : 16 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    ax = init_ax(fig, [0.2, 0.2, 0.6, 0.6])
    ax.set_xlim(mass_lim)
    ax.set_ylim(lha_lim)
    ax.set_xscale('log')

    ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][loew_indx]],
               numpy.ma.log10(dapall['SFR_1RE'][loew_indx])+41.27,
               color='0.5', marker='.', s=20, lw=0, zorder=3, alpha=0.3)

    sc = ax.scatter(drpall['nsa_sersic_mass'][dapall['DRPALLINDX'][hiew_indx]],
                    numpy.ma.log10(dapall['SFR_1RE'][hiew_indx])+41.27,
                    c=decrement[hiew_indx], vmin=2.8, vmax=7,
                    marker='.', s=20, lw=0, zorder=4, cmap='inferno')

    ax.text(-0.15, 0.5, r'$\log\ \langle L_{{\rm H}\alpha}\rangle_e$ [erg/s]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes,
            rotation='vertical')

    ax.text(0.5, -0.12, r'$\mathcal{M}_\ast$ [$h^{-2} \mathcal{M}_\odot$]',
            horizontalalignment='center', verticalalignment='center', transform=ax.transAxes)

    cax = fig.add_axes([0.58, 0.25, 0.20, 0.01])
    cb = pyplot.colorbar(sc, cax=cax, orientation='horizontal', ticks=[3, 4, 5, 6, 7])
    cax.text(0.5, 3.0, r'$L_{{\rm H}\alpha}/L_{{\rm H}\beta}$',
             ha='center', va='center', transform=cax.transAxes) #, fontsize=10)

    ofile = '../ms/figs/mass_lha.pdf'
    fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall = fits.open(dapall_file)['DAPALL'].data

    drpall_file = '/Users/westfall/Work/MaNGA/testing/dr15/drpall-v2_4_3.fits'
    drpall = fits.open(drpall_file)[1].data

    mass_lha(drpall, dapall)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))




r"""
Defines a :class:`~mangadap.util.bitmask.BitMask` objects for use in the DAP
based on the definition from the MaNGA DRP.


----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
from IPython import embed

from .bitmask import BitMask
from ..config import defaults


class DRPFitsBitMask(BitMask):
    r"""
    Structure with the DRP mask bits.

    The defined bits are listed at :ref:`metadatamodel-drp3pixmask`.

    Mode is not checked
    """
    def __init__(self, sdss_maskbits=None, mode='CUBE'):
        _sdss_maskbits = defaults.sdss_maskbits_file() if sdss_maskbits is None else sdss_maskbits
        tmp = BitMask.from_par_file(str(_sdss_maskbits),
                                    'MANGA_DRP3PIXMASK' if mode == 'CUBE' else 'MANGA_DRP2PIXMASK')
        keys, descr = tmp._init_objs()
        super(DRPFitsBitMask, self).__init__(keys, descr=descr)


class DRPQuality3DBitMask(BitMask):
    r"""
    Structure with the definition of the DRP3QUAL mask bits.

    The defined bits are listed at :ref:`metadatamodel-drp3qual`.
    """
    def __init__(self, sdss_maskbits=None):
        _sdss_maskbits = defaults.sdss_maskbits_file() if sdss_maskbits is None else sdss_maskbits
        tmp = BitMask.from_par_file(str(_sdss_maskbits), 'MANGA_DRP3QUAL')
        keys, descr = tmp._init_objs()
        super(DRPQuality3DBitMask, self).__init__(keys, descr=descr)


